import bmcs_utils.api as bu

from bmcs_shell.folding.geometry.wb_tessellation.wb_num_tessellation_base import WBNumTessellationBase
from bmcs_shell.folding.geometry.wb_tessellation.wb_num_tessellation import WBNumTessellation
from bmcs_shell.folding.geometry.wb_cell.wb_cell_5p_v2 import \
    WBCell5ParamV2
import traits.api as tr
import numpy as np
from numpy import cos, sin, sqrt
from scipy.optimize import minimize
import k3d
import random

class WBTessellation5PV2(WBNumTessellation):
    name = 'WBTessellation5PV2'

    wb_cell = bu.Instance(WBCell5ParamV2, ())
    tree = ['wb_cell']
    X_Ia = tr.DelegatesTo('wb_cell')
    I_Fi = tr.DelegatesTo('wb_cell')

    sigma_sol_num = bu.Int(-1, GEO=True)
    rho_sol_num = bu.Int(-1, GEO=True)

    new_sol = bu.Bool(True)

    # Note: Update traits to 6.3.2 in order for the following command to work!!
    @tr.observe('wb_cell.+GEO', post_init=True)
    def update_after_wb_cell_GEO_changes(self, event):
        self.event_geo = not self.event_geo
        self.update_plot(self.pb)

    ipw_view = bu.View(
        *WBNumTessellation.ipw_view.content,
        bu.Item('sigma_sol_num'),
        bu.Item('rho_sol_num'),
        bu.Item('new_sol')
    )

    sol = tr.Property(depends_on='+GEO')
    @tr.cached_property
    def _get_sol(self):
        # rhos, sigmas = self.get_3_cells_angles()
        # print('original sigmas=', sigmas)
        # print('original rhos=', rhos)
        # # rhos = 2 * np.pi - rhos
        # # sigmas = 2 * np.pi - sigmas
        # rhos = -rhos
        # sigmas = -sigmas
        # print('sigmas=', sigmas)
        # print('rhos=', rhos)
        #
        # if self.sigma_sol_num != -1 and self.rho_sol_num != -1:
        #     print('Solution {} was used.'.format(self.sigma_sol_num))
        #     sol = np.array([sigmas[self.sigma_sol_num - 1], rhos[self.rho_sol_num - 1]])
        #     return sol
        #
        # sigma_sol_idx = np.argmin(np.abs(sigmas - sol[0]))
        # rho_sol_idx = np.argmin(np.abs(rhos - sol[1]))
        #
        # if sigma_sol_idx != rho_sol_idx:
        #     print('Warning: sigma_sol_idx != rho_sol_idx, num solution is picked!')
        # else:
        #     diff = np.min(np.abs(sigmas - sol[0])) + np.min(np.abs(rhos - sol[1]))
        #     print('Solution {} was picked (nearst to numerical sol), diff={}'.format(sigma_sol_idx + 1, diff))
        #     sol = np.array([sigmas[sigma_sol_idx], rhos[rho_sol_idx]])

        # Solving with only 4th solution
        if self.new_sol:
            rhos, sigmas = self.get_3_cells_angles(sol_num=4)
        else:
            rhos, sigmas = self.get_3_cells_angles_old(sol_num=4)
        sol = np.array([sigmas[0], rhos[0]])
        print('Ana. solution:', sol)
        return sol

    def get_3_cells_angles_old(self, sol_num=None):
        a = self.wb_cell.a
        b = self.wb_cell.b
        c = self.wb_cell.c
        gamma = self.wb_cell.gamma
        beta = self.wb_cell.beta

        cos_psi1 = ((b ** 2 - a ** 2) - a * sqrt(a ** 2 + b ** 2) * cos(beta)) / (b * sqrt(a ** 2 + b ** 2) * sin(beta))
        sin_psi1 = sqrt(
            a ** 2 * (3 * b ** 2 - a ** 2) + 2 * a * (b ** 2 - a ** 2) * sqrt(a ** 2 + b ** 2) * cos(beta) - (
                    a ** 2 + b ** 2) ** 2 * cos(beta) ** 2) / (b * sqrt(a ** 2 + b ** 2) * sin(beta))
        cos_psi5 = (sqrt(a ** 2 + b ** 2) * cos(beta) - a * cos(2 * gamma)) / (b * sin(2 * gamma))
        sin_psi5 = sqrt(b ** 2 + 2 * a * sqrt(a ** 2 + b ** 2) * cos(beta) * cos(2 * gamma) - (a ** 2 + b ** 2) * (
                cos(beta) ** 2 + cos(2 * gamma) ** 2)) / (b * sin(2 * gamma))
        cos_psi6 = (a - sqrt(a ** 2 + b ** 2) * cos(beta) * cos(2 * gamma)) / (
                sqrt(a ** 2 + b ** 2) * sin(beta) * sin(2 * gamma))
        sin_psi6 = sqrt(b ** 2 + 2 * a * sqrt(a ** 2 + b ** 2) * cos(beta) * cos(2 * gamma) - (a ** 2 + b ** 2) * (
                cos(beta) ** 2 + cos(2 * gamma) ** 2)) / (sqrt(a ** 2 + b ** 2) * sin(beta) * sin(2 * gamma))
        cos_psi1plus6 = cos_psi1 * cos_psi6 - sin_psi1 * sin_psi6
        sin_psi1plus6 = sin_psi1 * cos_psi6 + cos_psi1 * sin_psi6

        sin_phi1 = sin_psi1plus6
        sin_phi2 = sin_psi5
        sin_phi3 = sin_psi5
        sin_phi4 = sin_psi1plus6

        cos_phi1 = cos_psi1plus6
        cos_phi2 = cos_psi5
        cos_phi3 = cos_psi5
        cos_phi4 = cos_psi1plus6

        # ALWAYS USE THIS TO FIND DEFINITE ANGLE IN [-pi, pi] from sin and cos
        phi1 = np.arctan2(sin_phi1, cos_phi1)
        phi2 = np.arctan2(sin_phi2, cos_phi2)
        phi3 = phi2
        phi4 = phi1

        W = 2 * a * (1 - cos(phi1 + phi3)) * (
                    sin(2 * gamma) ** 2 * a * c ** 2 - (cos(phi2) + cos(phi4)) * (cos(2 * gamma) - 1) * sin(
                2 * gamma) * b * c ** 2 - 2 * (cos(phi2) + cos(phi4)) * sin(2 * gamma) * a * b * c - 2 * a * c * (
                                2 * a - c) * (cos(2 * gamma) + 1) + 2 * a * b ** 2 * (cos(phi1 + phi3) + 1)) - (
                        ((a - c) ** 2 + b ** 2) * (cos(phi2) ** 2 + cos(phi4) ** 2) - 2 * cos(phi4) * cos(phi2) * (
                            (a - c) ** 2 + cos(phi1 + phi3) * b ** 2)) * c ** 2 * sin(2 * gamma) ** 2

        T_rho = ((a - c) ** 2 + b ** 2) * (cos(phi1 + phi3) - 1) * (
                    a * (2 * b ** 2 * (cos(phi1 + phi3) + 1) + 4 * a * (a - c)) * cos(2 * gamma)
                    + 2 * b * sin(2 * gamma) * (b ** 2 * cos(phi2) * (cos(phi1 + phi3) + 1) + (a - c) * (
                        (a - c) * cos(phi2) + a * cos(phi2) + c * cos(phi4))) + 2 * a * (
                                2 * a * (a - c) - b ** 2 * (cos(phi1 + phi3) + 1)))

        T_sigma = ((a - c) ** 2 + b ** 2) * (cos(phi1 + phi3) - 1) * (
                    a * (2 * b ** 2 * (cos(phi1 + phi3) + 1) + 4 * a * (a - c)) * cos(2 * gamma)
                    + 2 * b * sin(2 * gamma) * (b ** 2 * cos(phi4) * (cos(phi1 + phi3) + 1) + (a - c) * (
                        (a - c) * cos(phi4) + a * cos(phi4) + c * cos(phi2))) + 2 * a * (
                                2 * a * (a - c) - b ** 2 * (cos(phi1 + phi3) + 1)))
        # print('W:', W)
        # print('T_rho:', T_rho)
        # print('T_sigma:', T_sigma)

        rhos = []
        sigmas = []
        get_angle = lambda x: np.arctan(x) * 2

        if sol_num == 1 or sol_num is None:
            sol_P1_t_1 = (2 * b * (a - c) * (cos(phi1 + phi3) - 1) * sqrt((a - c) ** 2 + b ** 2) * sqrt(W) - 2 * b * (
                        a * b * c * sin(phi1 + phi3) * (cos(phi1 + phi3) - 1) * cos(2 * gamma) + (
                            ((a - c) ** 2 + b ** 2 * cos(phi1 + phi3)) * cos(phi2) - cos(phi4) * (
                                (a - c) ** 2 + b ** 2)) * sin(phi1 + phi3) * c * sin(2 * gamma) + a * b * sin(
                    phi1 + phi3) * (cos(phi1 + phi3) - 1) * (2 * a - c)) * sqrt((a - c) ** 2 + b ** 2)
                          + (cos(phi1 + phi3) - 1) * (2 * b ** 2 * cos(phi1 + phi3) + 4 * (a - c) ** 2 + 2 * b ** 2) * sqrt(
                        (a - c) ** 2 + b ** 2) * sqrt(
                        4 * a * b ** 2 * (cos(phi2) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) - (a ** 2 + b ** 2) * (
                                    cos(phi2) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) ** 2)) / (
                         (2 * b * ((a - c) ** 2 + b ** 2) * sin(phi1 + phi3) * sqrt(W) - T_rho))

            sol_Q1_u_1 = (2 * b * (a - c) * (cos(phi1 + phi3) - 1) * sqrt((a - c) ** 2 + b ** 2) * sqrt(W) - 2 * b * (
                    a * b * c * sin(phi1 + phi3) * (cos(phi1 + phi3) - 1) * cos(2 * gamma) + (
                    ((a - c) ** 2 + b ** 2 * cos(phi1 + phi3)) * cos(phi4) - cos(phi2) * (
                    (a - c) ** 2 + b ** 2)) * sin(phi1 + phi3) * c * sin(2 * gamma) + a * b * sin(
                phi1 + phi3) * (cos(phi1 + phi3) - 1) * (2 * a - c)) * sqrt((a - c) ** 2 + b ** 2)
                          + (cos(phi1 + phi3) - 1) * (
                                      2 * b ** 2 * cos(phi1 + phi3) + 4 * (a - c) ** 2 + 2 * b ** 2) * sqrt(
                        (a - c) ** 2 + b ** 2) * sqrt(
                        4 * a * b ** 2 * (cos(phi4) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) - (
                                    a ** 2 + b ** 2) * (
                                cos(phi4) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) ** 2)) / (
                             (2 * b * ((a - c) ** 2 + b ** 2) * sin(phi1 + phi3) * sqrt(W) - T_sigma))
            rhos.append(get_angle(sol_P1_t_1))
            sigmas.append(get_angle(sol_Q1_u_1))

        if sol_num == 2 or sol_num is None:
            sol_P1_t_2 = (2 * b * (a - c) * (cos(phi1 + phi3) - 1) * sqrt((a - c) ** 2 + b ** 2) * sqrt(W) - 2 * b * (
                        a * b * c * sin(phi1 + phi3) * (cos(phi1 + phi3) - 1) * cos(2 * gamma) + (
                            ((a - c) ** 2 + b ** 2 * cos(phi1 + phi3)) * cos(phi2) - cos(phi4) * (
                                (a - c) ** 2 + b ** 2)) * sin(phi1 + phi3) * c * sin(2 * gamma) + a * b * sin(
                    phi1 + phi3) * (cos(phi1 + phi3) - 1) * (2 * a - c)) * sqrt((a - c) ** 2 + b ** 2)
                          - (cos(phi1 + phi3) - 1) * (2 * b ** 2 * cos(phi1 + phi3) + 4 * (a - c) ** 2 + 2 * b ** 2) * sqrt(
                        (a - c) ** 2 + b ** 2) * sqrt(
                        4 * a * b ** 2 * (cos(phi2) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) - (a ** 2 + b ** 2) * (
                                    cos(phi2) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) ** 2)) / (
                         (2 * b * ((a - c) ** 2 + b ** 2) * sin(phi1 + phi3) * sqrt(W) - T_rho))

            sol_Q1_u_2 = (2 * b * (a - c) * (cos(phi1 + phi3) - 1) * sqrt((a - c) ** 2 + b ** 2) * sqrt(W) - 2 * b * (
                    a * b * c * sin(phi1 + phi3) * (cos(phi1 + phi3) - 1) * cos(2 * gamma) + (
                    ((a - c) ** 2 + b ** 2 * cos(phi1 + phi3)) * cos(phi4) - cos(phi2) * (
                    (a - c) ** 2 + b ** 2)) * sin(phi1 + phi3) * c * sin(2 * gamma) + a * b * sin(
                phi1 + phi3) * (cos(phi1 + phi3) - 1) * (2 * a - c)) * sqrt((a - c) ** 2 + b ** 2)
                          - (cos(phi1 + phi3) - 1) * (
                                      2 * b ** 2 * cos(phi1 + phi3) + 4 * (a - c) ** 2 + 2 * b ** 2) * sqrt(
                        (a - c) ** 2 + b ** 2) * sqrt(
                        4 * a * b ** 2 * (cos(phi4) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) - (
                                    a ** 2 + b ** 2) * (
                                cos(phi4) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) ** 2)) / (
                             (2 * b * ((a - c) ** 2 + b ** 2) * sin(phi1 + phi3) * sqrt(W) - T_sigma))
            rhos.append(get_angle(sol_P1_t_2))
            sigmas.append(get_angle(sol_Q1_u_2))

        if sol_num == 3 or sol_num is None:
            sol_P2_t_1 = (-2 * b * (a - c) * (cos(phi1 + phi3) - 1) * sqrt((a - c) ** 2 + b ** 2) * sqrt(W) - 2 * b * (
                        a * b * c * sin(phi1 + phi3) * (cos(phi1 + phi3) - 1) * cos(2 * gamma) + (
                            ((a - c) ** 2 + b ** 2 * cos(phi1 + phi3)) * cos(phi2) - cos(phi4) * (
                                (a - c) ** 2 + b ** 2)) * sin(phi1 + phi3) * c * sin(2 * gamma) + a * b * sin(
                    phi1 + phi3) * (cos(phi1 + phi3) - 1) * (2 * a - c)) * sqrt((a - c) ** 2 + b ** 2)
                          + (cos(phi1 + phi3) - 1) * (2 * b ** 2 * cos(phi1 + phi3) + 4 * (a - c) ** 2 + 2 * b ** 2) * sqrt(
                        (a - c) ** 2 + b ** 2) * sqrt(
                        4 * a * b ** 2 * (cos(phi2) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) - (a ** 2 + b ** 2) * (
                                    cos(phi2) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) ** 2)) / (
                         (-2 * b * ((a - c) ** 2 + b ** 2) * sin(phi1 + phi3) * sqrt(W) - T_rho))

            sol_Q2_u_1 = (-2 * b * (a - c) * (cos(phi1 + phi3) - 1) * sqrt((a - c) ** 2 + b ** 2) * sqrt(W) - 2 * b * (
                    a * b * c * sin(phi1 + phi3) * (cos(phi1 + phi3) - 1) * cos(2 * gamma) + (
                    ((a - c) ** 2 + b ** 2 * cos(phi1 + phi3)) * cos(phi4) - cos(phi2) * (
                    (a - c) ** 2 + b ** 2)) * sin(phi1 + phi3) * c * sin(2 * gamma) + a * b * sin(
                phi1 + phi3) * (cos(phi1 + phi3) - 1) * (2 * a - c)) * sqrt((a - c) ** 2 + b ** 2)
                          + (cos(phi1 + phi3) - 1) * (
                                      2 * b ** 2 * cos(phi1 + phi3) + 4 * (a - c) ** 2 + 2 * b ** 2) * sqrt(
                        (a - c) ** 2 + b ** 2) * sqrt(
                        4 * a * b ** 2 * (cos(phi4) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) - (
                                    a ** 2 + b ** 2) * (
                                cos(phi4) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) ** 2)) / (
                             (-2 * b * ((a - c) ** 2 + b ** 2) * sin(phi1 + phi3) * sqrt(W) - T_sigma))
            rhos.append(get_angle(sol_P2_t_1))
            sigmas.append(get_angle(sol_Q2_u_1))

        if sol_num == 4 or sol_num is None:
            sol_P2_t_2 = (-2 * b * (a - c) * (cos(phi1 + phi3) - 1) * sqrt((a - c) ** 2 + b ** 2) * sqrt(W) - 2 * b * (
                        a * b * c * sin(phi1 + phi3) * (cos(phi1 + phi3) - 1) * cos(2 * gamma) + (
                            ((a - c) ** 2 + b ** 2 * cos(phi1 + phi3)) * cos(phi2) - cos(phi4) * (
                                (a - c) ** 2 + b ** 2)) * sin(phi1 + phi3) * c * sin(2 * gamma) + a * b * sin(
                    phi1 + phi3) * (cos(phi1 + phi3) - 1) * (2 * a - c)) * sqrt((a - c) ** 2 + b ** 2)
                          - (cos(phi1 + phi3) - 1) * (2 * b ** 2 * cos(phi1 + phi3) + 4 * (a - c) ** 2 + 2 * b ** 2) * sqrt(
                        (a - c) ** 2 + b ** 2) * sqrt(
                        4 * a * b ** 2 * (cos(phi2) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) - (a ** 2 + b ** 2) * (
                                    cos(phi2) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) ** 2)) / (
                         (-2 * b * ((a - c) ** 2 + b ** 2) * sin(phi1 + phi3) * sqrt(W) - T_rho))

            sol_Q2_u_2 = (-2 * b * (a - c) * (cos(phi1 + phi3) - 1) * sqrt((a - c) ** 2 + b ** 2) * sqrt(W) - 2 * b * (
                        a * b * c * sin(phi1 + phi3) * (cos(phi1 + phi3) - 1) * cos(2 * gamma) + (
                            ((a - c) ** 2 + b ** 2 * cos(phi1 + phi3)) * cos(phi4) - cos(phi2) * (
                                (a - c) ** 2 + b ** 2)) * sin(phi1 + phi3) * c * sin(2 * gamma) + a * b * sin(
                    phi1 + phi3) * (cos(phi1 + phi3) - 1) * (2 * a - c)) * sqrt((a - c) ** 2 + b ** 2)
                          - (cos(phi1 + phi3) - 1) * (2 * b ** 2 * cos(phi1 + phi3) + 4 * (a - c) ** 2 + 2 * b ** 2) * sqrt(
                        (a - c) ** 2 + b ** 2) * sqrt(
                        4 * a * b ** 2 * (cos(phi4) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) - (a ** 2 + b ** 2) * (
                                    cos(phi4) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) ** 2)) / (
                         (-2 * b * ((a - c) ** 2 + b ** 2) * sin(phi1 + phi3) * sqrt(W) - T_sigma))
            rhos.append(get_angle(sol_P2_t_2))
            sigmas.append(get_angle(sol_Q2_u_2))

            # print('sol_P2_t_2:', sol_P2_t_2)
            # print('sol_Q2_u_2:', sol_Q2_u_2)

        # sol_P is tan(rho/2), sol_Q is tan(sigma/2)

        return -np.array(rhos), -np.array(sigmas)

    def get_3_cells_angles(self, sol_num=None):
        a = self.wb_cell.a
        b = self.wb_cell.b
        c = self.wb_cell.c
        gamma = self.wb_cell.gamma
        beta = self.wb_cell.beta

        cos_psi1 = ((b ** 2 - a ** 2) - a * sqrt(a ** 2 + b ** 2) * cos(beta)) / (b * sqrt(a ** 2 + b ** 2) * sin(beta))
        sin_psi1 = sqrt(
            a ** 2 * (3 * b ** 2 - a ** 2) + 2 * a * (b ** 2 - a ** 2) * sqrt(a ** 2 + b ** 2) * cos(beta) - (
                    a ** 2 + b ** 2) ** 2 * cos(beta) ** 2) / (b * sqrt(a ** 2 + b ** 2) * sin(beta))
        cos_psi5 = (sqrt(a ** 2 + b ** 2) * cos(beta) - a * cos(2 * gamma)) / (b * sin(2 * gamma))
        sin_psi5 = sqrt(b ** 2 + 2 * a * sqrt(a ** 2 + b ** 2) * cos(beta) * cos(2 * gamma) - (a ** 2 + b ** 2) * (
                cos(beta) ** 2 + cos(2 * gamma) ** 2)) / (b * sin(2 * gamma))
        cos_psi6 = (a - sqrt(a ** 2 + b ** 2) * cos(beta) * cos(2 * gamma)) / (
                sqrt(a ** 2 + b ** 2) * sin(beta) * sin(2 * gamma))
        sin_psi6 = sqrt(b ** 2 + 2 * a * sqrt(a ** 2 + b ** 2) * cos(beta) * cos(2 * gamma) - (a ** 2 + b ** 2) * (
                cos(beta) ** 2 + cos(2 * gamma) ** 2)) / (sqrt(a ** 2 + b ** 2) * sin(beta) * sin(2 * gamma))
        cos_psi1plus6 = cos_psi1 * cos_psi6 - sin_psi1 * sin_psi6
        sin_psi1plus6 = sin_psi1 * cos_psi6 + cos_psi1 * sin_psi6

        sin_phi1 = sin_psi1plus6
        sin_phi2 = sin_psi5
        sin_phi3 = sin_psi5
        sin_phi4 = sin_psi1plus6

        cos_phi1 = cos_psi1plus6
        cos_phi2 = cos_psi5
        cos_phi3 = cos_psi5
        cos_phi4 = cos_psi1plus6

        # ALWAYS USE THIS TO FIND DEFINITE ANGLE IN [-pi, pi] from sin and cos
        phi1 = np.arctan2(sin_phi1, cos_phi1)
        phi2 = np.arctan2(sin_phi2, cos_phi2)
        phi3 = phi2
        phi4 = phi1

        e = sqrt((a - c) ** 2 + b ** 2)
        f = sqrt((a - c) ** 2 + b ** 2 * cos(phi1 + phi3))
        W = 2 * (1 - cos(phi1 + phi3)) * (
                    (a ** 2 - cos(phi2) * cos(phi4) * b ** 2) * c ** 2 * sin(2 * gamma) ** 2 - (
                        cos(phi2) + cos(phi4)) * ((cos(2 * gamma) - 1) * c + 2 * a) * sin(
                2 * gamma) * a * b * c - 2 * a ** 2 * c * (2 * a - c) * (cos(2 * gamma) + 1) + 2 * a ** 2 * b ** 2 * (
                                cos(phi1 + phi3) + 1)) - e ** 2 * (cos(phi2) - cos(phi4)) ** 2 * c ** 2 * sin(2 * gamma) ** 2

        T_rho = (cos(phi1 + phi3) - 1) * (
                    a * (b ** 2 * (cos(phi1 + phi3) + 1) + 2 * a * (a - c)) * (cos(2 * gamma) + 1)
                    + b * (((a - c) ** 2 + b ** 2 + (a - c) ** 2 + cos(phi1 + phi3) * b ** 2) * cos(phi2) + (
                        a - c) * c * (cos(phi2) + cos(phi4))) * sin(2 * gamma) - 2 * a * b ** 2 * (
                                cos(phi1 + phi3) + 1))

        T_sigma = (cos(phi1 + phi3) - 1) * (
                    a * (b ** 2 * (cos(phi1 + phi3) + 1) + 2 * a * (a - c)) * (cos(2 * gamma) + 1)
                    + b * (((a - c) ** 2 + b ** 2 + (a - c) ** 2 + cos(phi1 + phi3) * b ** 2) * cos(phi4) + (
                        a - c) * c * (cos(phi2) + cos(phi4))) * sin(2 * gamma) - 2 * a * b ** 2 * (
                                cos(phi1 + phi3) + 1))
        # print('W:', W)
        # print('T_rho:', T_rho)
        # print('T_sigma:', T_sigma)

        rhos = []
        sigmas = []
        get_angle = lambda x: np.arctan(x) * 2

        if sol_num == 1 or sol_num is None:
            sol_P1_t_1 = (b * (a - c) * (cos(phi1 + phi3) - 1) * sqrt(W) - b * (
                    a * b * c * sin(phi1 + phi3) * (cos(phi1 + phi3) - 1) * cos(2 * gamma) + (
                    cos(phi2) * f ** 2 - cos(phi4) * e ** 2) * sin(phi1 + phi3) * c * sin(
                2 * gamma) + a * b * sin(phi1 + phi3) * (cos(phi1 + phi3) - 1) * (2 * a - c))
                          + (cos(phi1 + phi3) - 1) * (e ** 2 + f ** 2) * sqrt(
                        4 * a * b ** 2 * (cos(phi2) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) - (
                                a ** 2 + b ** 2) * (
                                cos(phi2) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) ** 2)) / (
                                     e * (b * sin(phi1 + phi3) * sqrt(W) - T_rho))

            sol_Q1_t_1 = (b * (a - c) * (cos(phi1 + phi3) - 1) * sqrt(W) - b * (
                    a * b * c * sin(phi1 + phi3) * (cos(phi1 + phi3) - 1) * cos(2 * gamma) + (
                    cos(phi4) * f ** 2 - cos(phi2) * e ** 2) * sin(phi1 + phi3) * c * sin(
                2 * gamma) + a * b * sin(phi1 + phi3) * (cos(phi1 + phi3) - 1) * (2 * a - c))
                          + (cos(phi1 + phi3) - 1) * (e ** 2 + f ** 2) * sqrt(
                        4 * a * b ** 2 * (cos(phi2) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) - (
                                a ** 2 + b ** 2) * (
                                cos(phi2) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) ** 2)) / (
                                     e * (b * sin(phi1 + phi3) * sqrt(W) - T_sigma))

            rhos.append(get_angle(sol_P1_t_1))
            sigmas.append(get_angle(sol_Q1_t_1))

        if sol_num == 2 or sol_num is None:
            sol_P1_t_2 = (b * (a - c) * (cos(phi1 + phi3) - 1) * sqrt(W) - b * (
                    a * b * c * sin(phi1 + phi3) * (cos(phi1 + phi3) - 1) * cos(2 * gamma) + (
                    cos(phi2) * f ** 2 - cos(phi4) * e ** 2) * sin(phi1 + phi3) * c * sin(
                2 * gamma) + a * b * sin(phi1 + phi3) * (cos(phi1 + phi3) - 1) * (2 * a - c))
                          - (cos(phi1 + phi3) - 1) * (e ** 2 + f ** 2) * sqrt(
                        4 * a * b ** 2 * (cos(phi2) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) - (
                                a ** 2 + b ** 2) * (
                                cos(phi2) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) ** 2)) / (
                                     e * (b * sin(phi1 + phi3) * sqrt(W) - T_rho))

            sol_Q1_t_2 = (b * (a - c) * (cos(phi1 + phi3) - 1) * sqrt(W) - b * (
                    a * b * c * sin(phi1 + phi3) * (cos(phi1 + phi3) - 1) * cos(2 * gamma) + (
                    cos(phi4) * f ** 2 - cos(phi2) * e ** 2) * sin(phi1 + phi3) * c * sin(
                2 * gamma) + a * b * sin(phi1 + phi3) * (cos(phi1 + phi3) - 1) * (2 * a - c))
                          - (cos(phi1 + phi3) - 1) * (e ** 2 + f ** 2) * sqrt(
                        4 * a * b ** 2 * (cos(phi2) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) - (
                                a ** 2 + b ** 2) * (
                                cos(phi2) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) ** 2)) / (
                                     e * (b * sin(phi1 + phi3) * sqrt(W) - T_sigma))
            rhos.append(get_angle(sol_P1_t_2))
            sigmas.append(get_angle(sol_Q1_t_2))

        if sol_num == 3 or sol_num is None:
            sol_P2_t_1 = (-b * (a - c) * (cos(phi1 + phi3) - 1) * sqrt(W) - b * (
                    a * b * c * sin(phi1 + phi3) * (cos(phi1 + phi3) - 1) * cos(2 * gamma) + (
                    cos(phi2) * f ** 2 - cos(phi4) * e ** 2) * sin(phi1 + phi3) * c * sin(
                2 * gamma) + a * b * sin(phi1 + phi3) * (cos(phi1 + phi3) - 1) * (2 * a - c))
                          + (cos(phi1 + phi3) - 1) * (e ** 2 + f ** 2) * sqrt(
                        4 * a * b ** 2 * (cos(phi2) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) - (
                                a ** 2 + b ** 2) * (
                                cos(phi2) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) ** 2)) / (
                                     e * (-b * sin(phi1 + phi3) * sqrt(W) - T_rho))

            sol_Q2_t_1 = (-b * (a - c) * (cos(phi1 + phi3) - 1) * sqrt(W) - b * (
                    a * b * c * sin(phi1 + phi3) * (cos(phi1 + phi3) - 1) * cos(2 * gamma) + (
                    cos(phi4) * f ** 2 - cos(phi2) * e ** 2) * sin(phi1 + phi3) * c * sin(
                2 * gamma) + a * b * sin(phi1 + phi3) * (cos(phi1 + phi3) - 1) * (2 * a - c))
                          + (cos(phi1 + phi3) - 1) * (e ** 2 + f ** 2) * sqrt(
                        4 * a * b ** 2 * (cos(phi2) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) - (
                                a ** 2 + b ** 2) * (
                                cos(phi2) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) ** 2)) / (
                                     e * (-b * sin(phi1 + phi3) * sqrt(W) - T_sigma))

            rhos.append(get_angle(sol_P2_t_1))
            sigmas.append(get_angle(sol_Q2_t_1))

        if sol_num == 4 or sol_num is None:
            sol_P2_t_2 = (-b * (a - c) * (cos(phi1 + phi3) - 1) * sqrt(W) - b * (
                    a * b * c * sin(phi1 + phi3) * (cos(phi1 + phi3) - 1) * cos(2 * gamma) + (
                    cos(phi2) * f ** 2 - cos(phi4) * e ** 2) * sin(phi1 + phi3) * c * sin(
                2 * gamma) + a * b * sin(phi1 + phi3) * (cos(phi1 + phi3) - 1) * (2 * a - c))
                          - (cos(phi1 + phi3) - 1) * (e ** 2 + f ** 2) * sqrt(
                        4 * a * b ** 2 * (cos(phi2) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) - (
                                a ** 2 + b ** 2) * (
                                cos(phi2) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) ** 2)) / (
                                     e * (-b * sin(phi1 + phi3) * sqrt(W) - T_rho))

            sol_Q2_t_2 = (-b * (a - c) * (cos(phi1 + phi3) - 1) * sqrt(W) - b * (
                    a * b * c * sin(phi1 + phi3) * (cos(phi1 + phi3) - 1) * cos(2 * gamma) + (
                    cos(phi4) * f ** 2 - cos(phi2) * e ** 2) * sin(phi1 + phi3) * c * sin(
                2 * gamma) + a * b * sin(phi1 + phi3) * (cos(phi1 + phi3) - 1) * (2 * a - c))
                          - (cos(phi1 + phi3) - 1) * (e ** 2 + f ** 2) * sqrt(
                        4 * a * b ** 2 * (cos(phi2) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) - (
                                a ** 2 + b ** 2) * (
                                cos(phi2) * sin(2 * gamma) * b + (cos(2 * gamma) + 1) * a) ** 2)) / (
                                     e * (-b * sin(phi1 + phi3) * sqrt(W) - T_sigma))
            print('sol_P2_t_2:', sol_P2_t_2)
            print('sol_Q2_t_2:', sol_Q2_t_2)
            rhos.append(get_angle(sol_P2_t_2))
            sigmas.append(get_angle(sol_Q2_t_2))

        # sol_P is tan(rho/2), sol_Q is tan(sigma/2)

        return -np.array(rhos), -np.array(sigmas)


import bmcs_utils.api as bu
import numpy as np
import traits.api as tr
import k3d

class WBCell(bu.Model):
    name = 'Waterbomb cell'

    plot_backend = 'k3d'

    K3D_NODES_LABELS = 'nodes_labels'
    K3D_WIREFRAME = 'wireframe'
    K3D_CELL_MESH = 'cell_mesh'

    show_base_cell_ui = bu.Bool(True)
    show_node_labels = bu.Bool(False, GEO=True)
    show_wireframe = bu.Bool(True, GEO=True)
    opacity = bu.Float(0.6, GEO=True)

    ipw_view = bu.View(
        bu.Item('show_node_labels'),
        bu.Item('show_wireframe'),
    ) if show_base_cell_ui else bu.View()

    X_Ia = tr.Property(depends_on='+GEO')
    '''Array with nodal coordinates I - node, a - dimension
    '''
    @tr.cached_property
    def _get_X_Ia(self):
        return np.array([[0., 0., 0.],
                         [1000., 930.99634691, 365.02849483],
                         [-1000., 930.99634691, 365.02849483],
                         [1000., -930.99634691, 365.02849483],
                         [-1000., -930.99634691, 365.02849483],
                         [764.84218728, 0., 644.21768724],
                         [-764.84218728, 0., 644.21768724]])

    I_Fi = tr.Property
    '''Triangle mapping '''
    @tr.cached_property
    def _get_I_Fi(self):
        return np.array([[0, 1, 2], [0, 3, 4], [0, 1, 5], [0, 5, 3], [0, 2, 6], [0, 6, 4]]).astype(np.int32)

    def setup_plot(self, pb):
        X_Ia = self.X_Ia.astype(np.float32)
        I_Fi = self.I_Fi.astype(np.uint32)
        cell_mesh = k3d.mesh(X_Ia, I_Fi,
                                opacity=self.opacity,
                                color=0x999999,
                                side='double')
        pb.plot_fig += cell_mesh
        pb.objects[self.K3D_CELL_MESH] = cell_mesh

        if self.show_wireframe:
            self._add_wireframe_to_fig(pb, X_Ia, I_Fi)
        if self.show_node_labels:
            self._add_nodes_labels_to_fig(pb, self.X_Ia)

    def update_plot(self, pb):
        # If cell interface was embedded in higher class, this method will be called when user changes parameters
        #  However, cell mesh object will not be there because setup_plot was not called
        if self.K3D_CELL_MESH in pb.objects:
            X_Ia = self.X_Ia.astype(np.float32)
            I_Fi = self.I_Fi.astype(np.uint32)
            cell_mesh = pb.objects[self.K3D_CELL_MESH]
            cell_mesh.vertices = X_Ia
            cell_mesh.indices = I_Fi
            cell_mesh.attributes = X_Ia[:, 2]

            if self.show_wireframe:
                if self.K3D_WIREFRAME in pb.objects:
                    wireframe = pb.objects[self.K3D_WIREFRAME]
                    wireframe.vertices = X_Ia
                    wireframe.indices = I_Fi
                else:
                    self._add_wireframe_to_fig(pb, X_Ia, I_Fi)
            else:
                if self.K3D_WIREFRAME in pb.objects:
                        pb.clear_object(self.K3D_WIREFRAME)

            if self.show_node_labels:
                if self.K3D_NODES_LABELS in pb.objects:
                    pb.clear_object(self.K3D_NODES_LABELS)
                self._add_nodes_labels_to_fig(pb, self.X_Ia)
            else:
                if self.K3D_NODES_LABELS in pb.objects:
                    pb.clear_object(self.K3D_NODES_LABELS)

    def _add_wireframe_to_fig(self, pb, X_Ia, I_Fi):
        k3d_mesh_wireframe = k3d.mesh(X_Ia,
                                      I_Fi,
                                      color=0x000000,
                                      wireframe=True)
        pb.plot_fig += k3d_mesh_wireframe
        pb.objects[self.K3D_WIREFRAME] = k3d_mesh_wireframe

    def _add_nodes_labels_to_fig(self, pb, X_Ia):
        text_list = []
        for I, X_a in enumerate(X_Ia):
            k3d_text = k3d.text('%g' % I, tuple(X_a), label_box=False, size=0.8, color=0x00FF00)
            pb.plot_fig += k3d_text
            text_list.append(k3d_text)
        pb.objects[self.K3D_NODES_LABELS] = text_list



# Indices reference:
# ------------------
# m: num of Gauss points
# r: num of curvilinear (natural) coordinates (r, s, t)
# i: num of shape functions max = 3, (h1, h2, h3)
# i: num of nodes in each element max = 3, (0, 1, 2)
# F: global num of elements?
# E: global num of elements?
# a: coordinates (x, y, z)
# b: coordinates (x, y, z)
# c: coordinates (x, y, z)
# d: coordinates (x, y, z)

import traits.api as tr
from ibvpy.mathkit.tensor import DELTA23_ab
import numpy as np

from bmcs_shell.folding.analysis.fets2d_mitc import FETS2DMITC
from bmcs_shell.folding.utils.vector_acos import \
    get_theta, get_theta_du
import k3d
from ibvpy.mathkit.linalg.sys_mtx_assembly import SysMtxArray
import bmcs_utils.api as bu

INPUT = '+cp_input'


# Kronecker delta
DELTA = np.zeros((3, 3,), dtype='f')
DELTA[(0, 1, 2), (0, 1, 2)] = 1

# Levi Civita symbol
EPS = np.zeros((3, 3, 3), dtype='f')
EPS[(0, 1, 2), (1, 2, 0), (2, 0, 1)] = 1
EPS[(2, 1, 0), (1, 0, 2), (0, 2, 1)] = -1

from bmcs_shell.folding.analysis.fem.fe_triangular_mesh import FETriangularMesh
from bmcs_shell.folding.analysis.fem.xdomain_fe_grid import XDomainFE

class TriXDomainMITC(XDomainFE):
    name = 'TriXDomainFE'
    '''
    Finite element discretization with dofs and mappings derived from the FE definition
    '''

    mesh = tr.Instance(FETriangularMesh)
    def _mesh_default(self):
        return FETriangularMesh(fets=FETS2DMITC())

    fets = tr.DelegatesTo('mesh')

    tree = ['mesh']

    change = tr.Event(GEO=True)

    plot_backend = 'k3d'

    n_dofs = tr.Property

    def _get_n_dofs(self):
        return len(self.mesh.X_Id) * self.mesh.n_nodal_dofs

    eta_w = tr.Property
    r'''Weight factors for numerical integration.
    '''

    def _get_eta_w(self):
        return self.fets.w_m

    Na_deta = tr.Property()
    r'''Derivatives of the shape functions in the integration points.
    '''
    @tr.cached_property
    def _get_Na_deta(self):
        return np.einsum('imr->mri', self.fets.dh_imr)

    x_0 = tr.Property()
    @tr.cached_property
    def _get_x_0(self):
        return self.mesh.X_Id

    x = tr.Property()
    @tr.cached_property
    def _get_x(self):
        return self.x_0

    F = tr.Property()
    @tr.cached_property
    def _get_F(self):
        return self.mesh.I_Fi

    T_Fab = tr.Property(depends_on='+GEO')
    @tr.cached_property
    def _get_T_Fab(self):
        return self.F_L_bases[:, 0, :]

    I_Ei = tr.Property
    def _get_I_Ei(self):
        return self.F_N

    x_Eia = tr.Property(depends_on='+GEO')
    @tr.cached_property
    def _get_x_Eia(self):
        X_Eia = self.X_Id[self.I_Ei, :]
        X_E0a = X_Eia[:, 0, :]
        X_Eia -= X_E0a[:, np.newaxis, :]
        X_Eic = np.einsum('Eac,Eic->Eia', self.T_Fab, X_Eia)
        return X_Eic[...]

    def U2u(self, U_Eia):
        u0_Eia = U_Eia[...,:-1]
        return u0_Eia

    def xU2u(self, U_Eio):
        u1_Eia = np.einsum('Eab,Eib->Eia', self.T_Fab, U_Eio[..., :-2])
        # u2_Eie =  np.einsum('ea,Eia->Eie', DELTA23_ab, u1_Eia)
        # return u2_Eie

        U1_Eio = np.copy(U_Eio).astype(np.float_)
        U1_Eio[..., :3, :3] = u1_Eia

        return U1_Eio

    def f2F(self, f_Eid):
        F0_Eia = np.concatenate( [f_Eid, np.zeros_like(f_Eid[...,:1])], axis=-1)
        return F0_Eia

    def xf2F(self, f_Eid):
        # F1_Eia = np.einsum('da,Eid->Eia', DELTA23_ab, f_Eid)

        print('f_Eid=', f_Eid)
        # F2_Eia = np.einsum('Eab,Eia->Eib', self.T_Fab, f_Eid)
        F2_Eia = np.einsum('Eab,Eia->Eib', self.T_Fab, f_Eid[..., :-2])
        f1_Eid = np.copy(f_Eid).astype(np.float_)
        f1_Eid[..., :-2] = F2_Eia

        return f1_Eid

    def k2K(self, K_Eiejf):
        K0_Eicjf = np.concatenate([K_Eiejf, np.zeros_like(K_Eiejf[:,:,:1,:,:])], axis=2)
        K0_Eicjd = np.concatenate([K0_Eicjf, np.zeros_like(K0_Eicjf[:,:,:,:,:1])], axis=4)
        return K0_Eicjd

    def xk2K(self, K_Eiejf):
        # K1_Eiejf = np.einsum('ea,fb,Eiejf->Eiajb', DELTA23_ab, DELTA23_ab, K_Eiejf) # correct
        T_Eeafb = np.einsum('Eea,Efb->Eeafb', self.T_Fab, self.T_Fab)
        #K_Eab = np.einsum('Eeafb,ef->Eab', T_Eeafb, k_ef)



        # K2_Eiajb = np.einsum('Eeafb,Eiejf->Eiajb', T_Eeafb, K_Eiejf)

        K2_Eiajb = np.einsum('Eeafb,Eiejf->Eiajb', T_Eeafb, K_Eiejf[:,:,:-2,:,:-2])
        K1_Eiejf = np.copy(K_Eiejf).astype(np.float_)
        K1_Eiejf[:,:,:-2,:,:-2] = K2_Eiajb

        #K2_Eicjd = np.einsum('Eca,Ebd,Eiajb->Eicjd', self.T_Fab, self.T_Fab, K1_Eicjd)
        #K2_Eicjd = np.einsum('Eac,Edb,Eiajb->Eicjd', self.T_Fab, self.T_Fab, K1_Eicjd)
        return K1_Eiejf

    I_CDij = tr.Property
    def _get_I_CDij(self):
        return self.mesh.I_CDij

    bc_J_F_xyz= tr.Property(depends_on='state_changed')
    @tr.cached_property
    def _get_bc_J_F_xyz(self):
        ix2 = int((self.mesh.n_phi_plus) / 2)
        F_I = self.I_CDij[ix2, :, 0, :].flatten()
        _, idx_remap = self.mesh.unique_node_map
        return idx_remap[F_I]

    bc_J_xyz = tr.Property(depends_on='state_changed')
    @tr.cached_property
    def _get_bc_J_xyz(self):
        I_M = self.I_CDij[(0, -1), :, (0, -1), :].flatten()
        _, idx_remap = self.mesh.unique_node_map
        J_M = idx_remap[I_M]
        return J_M

    bc_J_x = tr.Property(depends_on='state_changed')
    @tr.cached_property
    def _get_bc_J_x(self):
        I_M = self.I_CDij[:, (0, -1), :, (0, -1)].flatten()
        _, idx_remap = self.mesh.unique_node_map
        J_M = idx_remap[I_M]
        return J_M

    def setup_plot(self, pb):
        X_Ia = self.mesh.X_Ia.astype(np.float32)
        I_Fi = self.mesh.I_Fi.astype(np.uint32)

        X_Ma = X_Ia[self.bc_J_xyz]
        self.k3d_fixed_xyz = k3d.points(X_Ma)
        pb.plot_fig += self.k3d_fixed_xyz

        X_Ma = X_Ia[self.bc_J_x]
        self.k3d_fixed_x = k3d.points(X_Ma, color=0x22ffff)
        pb.plot_fig += self.k3d_fixed_x

        X_Ma = X_Ia[self.bc_J_F_xyz]
        self.k3d_load_z = k3d.points(X_Ma, color=0xff22ff)
        pb.plot_fig += self.k3d_load_z

        self.k3d_mesh = k3d.mesh(X_Ia,
                                 I_Fi,
                                 color=0x999999,
                                 side='double')
        pb.plot_fig += self.k3d_mesh

    def update_plot(self, pb):
        X_Ia = self.mesh.X_Ia.astype(np.float32)
        I_Fi = self.mesh.I_Fi.astype(np.uint32)

        self.k3d_fixed_xyz.positions = X_Ia[self.bc_J_xyz]
        self.k3d_fixed_x.positions = X_Ia[self.bc_J_x]
        self.k3d_load_z.positions = X_Ia[self.bc_J_F_xyz]

        mesh = self.k3d_mesh
        mesh.vertices = X_Ia
        mesh.indices = I_Fi

    def _get_du_dr_Fmra(self, U_o):
        dh_imr = self.fets.dh_imr
        dht_imr = self.fets.dht_imr
        _, v1_Fid, v2_Fid = self.v_vectors
        a = self.fets.a

        # Calculating du_dr
        U_o = np.arange(3 * 5) # TODO U_o comes from function arguments, this is just for testing
        nodes_num = self.mesh.X_Id.shape[0]
        U_Ie = np.reshape(U_o, (nodes_num, self.fets.n_nodal_dofs))
        U_Fie = U_Ie[self.mesh.I_Fi]
        disp_U_Fia = U_Fie[..., :3]
        rot_U_Fib = U_Fie[..., 3:]
        du_dr1_Fmria = np.einsum('Fia, imr ->Fmria', disp_U_Fia, dh_imr)
        du_dr1_Fmra = np.sum(du_dr1_Fmria, axis=3)

        alpha_idx = 0
        beta_idx = 1
        alpha_Fi1 = rot_U_Fib[..., alpha_idx, np.newaxis]
        beta_Fi1 = rot_U_Fib[..., beta_idx, np.newaxis]
        v2_alpha_Fid = v2_Fid * alpha_Fi1
        v1_beta_Fid = v1_Fid * beta_Fi1
        v1_v2_dif_Fid = v1_beta_Fid - v2_alpha_Fid
        du_dr2_Fmria = np.einsum('Fia, imr ->Fmria', 0.5 * a * v1_v2_dif_Fid, dht_imr)
        du_dr2_Fmra = np.sum(du_dr2_Fmria, axis=3)
        du_dr_Fmra = du_dr1_Fmra + du_dr2_Fmra
        return du_dr_Fmra

    v_vectors = tr.Property(depends_on='+GEO')
    @tr.cached_property
    def _get_v_vectors(self):
        # See P. 472 FEM by Zienkiewicz (ISBN: 1856176339)
        # Calculating v_n (director vector)
        n_Fd = self.norm_F_normals
        el_nodes_num = 3
        Vn_Fid = np.tile(n_Fd, (1, 1, el_nodes_num)).reshape(n_Fd.shape[0], n_Fd.shape[1], el_nodes_num)

        # Calculating v1 and v2 (vectors perpendicular to director vector)
        min_Fi1 = np.abs(Vn_Fid).argmin(axis=2)
        min_Fi1 = min_Fi1[..., np.newaxis]
        tmp_Fid = np.zeros_like(Vn_Fid, dtype=np.int_)
        tmp_Fid[..., :] = np.arange(3)
        min_mask_Fid = tmp_Fid == min_Fi1
        e_x_min_Fid = min_mask_Fid * 1
        V1_Fid = np.cross(e_x_min_Fid, Vn_Fid)
        V2_Fid = np.cross(Vn_Fid, V1_Fid)
        v1_Fid = self._normalize(V1_Fid)
        v2_Fid = self._normalize(V2_Fid)

        return Vn_Fid, v1_Fid, v2_Fid

    def _normalize(self, V_Fid):
        mag_n = np.sqrt(np.einsum('...i,...i', V_Fid, V_Fid))
        v_Fid = V_Fid / mag_n[:, np.newaxis]
        return v_Fid

    dx_dr_Fmrd = tr.Property(depends_on='+GEO')
    @tr.cached_property
    def _get_dx_dr_Fmrd(self):
        Vn_Fid, _, _ = self.v_vectors

        dh_imr = self.fets.dh_imr
        dht_imr = self.fets.dht_imr

        # X_Fid = self.X_Id[self.mesh.I_Fi]
        # X_Fid = self.X_Id[self.F_N]

        # get coords transformed to local
        X_Fid = self.x_Eia

        a = self.fets.a  # thickness (TODO, make it available as input)
        dx_dr1_Fmrid = np.einsum('Fid, imr -> Fmrid', X_Fid, dh_imr)
        dx_dr2_Fmrid = np.einsum('Fid, imr -> Fmrid', 0.5 * a * Vn_Fid, dht_imr)

        dx_dr1_Fmrd = np.sum(dx_dr1_Fmrid, axis=3)
        dx_dr2_Fmrd = np.sum(dx_dr2_Fmrid, axis=3)
        dx_dr_Fmrd = dx_dr1_Fmrd + dx_dr2_Fmrd
        return dx_dr_Fmrd

    B_Emiabo = tr.Property(depends_on='+GEO')
    @tr.cached_property
    def _get_B_Emiabo(self):
        delta35_co = np.zeros((3, 5), dtype='f')
        delta35_co[(0, 1, 2), (0, 1, 2)] = 1
        delta25_vo = np.zeros((2, 5), dtype='f')
        delta25_vo[(0, 1), (3, 4)] = 1

        _, v1_Fid, v2_Fid = self.v_vectors
        V_Ficv = np.zeros((*v2_Fid.shape, 2), dtype='f')
        V_Ficv[..., 0] = v1_Fid
        V_Ficv[..., 1] = v2_Fid

        # Thickness a
        a = self.fets.a
        dN_imr = self.fets.dh_imr
        dNt_imr = self.fets.dht_imr

        delta = np.identity(3)
        Diff1_abcd = 0.5 * (
                np.einsum('ac,bd->abcd', delta, delta) +
                np.einsum('ad,bc->abcd', delta, delta)
        )

        J_Fmrd = self.dx_dr_Fmrd

        inv_J_Fmrd = np.linalg.inv(J_Fmrd)
        det_J_Fm = np.linalg.det(J_Fmrd)
        B1_Emiabo = np.einsum('abcd, imr, co, Emrd -> Emiabo', Diff1_abcd, dN_imr, delta35_co, inv_J_Fmrd)
        B2_Emiabo = np.einsum('abcd, imr, Ficv, vo, Emrd -> Emiabo', Diff1_abcd, dNt_imr, 0.5 * a * V_Ficv, delta25_vo,
                              inv_J_Fmrd)
        B_Emiabo = B1_Emiabo + B2_Emiabo

        # B_Emiabo = np.flip(B_Emiabo, 2)

        return B_Emiabo, det_J_Fm

    def _get_B_Empf(self):
        B_Emiabo, _ = self.B_Emiabo
        # Mapping ab to p (3x3 -> 5)
        B_Emipo = B_Emiabo[:, :, :, (0, 1, 0, 1, 2), (0, 1, 1, 2, 0), :]

        E, m, i, p, o = B_Emipo.shape
        B_Empio = np.einsum('Emipo->Empio', B_Emipo)
        B_Empf = B_Empio.reshape((E, m, p, 3 * o))
        # p: index with max value 5
        # f: index with max value 15
        return B_Empf

    def map_U_to_field(self, U_o):
        # print('map_U_to_field')
        # # For testing with one element:
        # U_io = np.array([[1, 0, 0, 0, 0],
        #                  [0, 0, 0.5, 0, 0],
        #                  [0, 1, 0, 0, 0]], dtype=np.float_)
        print('U_o:', U_o)

        # TODO: check if the transformation caused by following line is needed
        U_Eio = U_o[self.o_Eia]
        # print('U_Eio,', U_Eio)
        # U_Eio = U_o.reshape((-1, self.fets.n_nodal_dofs))[self.mesh.I_Fi]
        # U_Eio = U_o.reshape((-1, self.fets.n_nodal_dofs))[self.F_N]

        # transform to local
        U_Eio = self.xU2u(U_Eio)

        print('U_Eio,', U_Eio)

        B_Emiabo, _ = self.B_Emiabo

        eps_Emab = np.einsum('Emiabo, Eio -> Emab', B_Emiabo, U_Eio)

        eps_Emp = eps_Emab[:, :, (0, 1, 0, 1, 0), (0, 1, 1, 2, 2)]
        return eps_Emp

    def map_field_to_F(self, sig_Ems):
        # print('map_field_to_F')
        print('sig_Es', sig_Ems)

        _, det_J_Fm = self.B_Emiabo

        B_Empf = self._get_B_Empf()
        f_Emf = self.integ_factor * np.einsum('m, Emsf, Ems, Em -> Emf', self.fets.w_m, B_Empf, sig_Ems, det_J_Fm)
        f_Ef = np.sum(f_Emf, axis=1)

        # o_Ei = self.o_Eia.reshape(-1, 3 * 5)
        # print('o_Ei:', o_Ei)
        # # o_Ei = self.o_Ia[self.mesh.I_Fi].reshape(-1, 3 * self.fets.n_nodal_dofs)
        # # o_Ei = self.o_Ia[self.F_N].reshape(-1, 3 * self.fets.n_nodal_dofs)
        # return o_Ei.flatten(), f_Ef.flatten()


        f_Eic = f_Ef.reshape(-1, 3, 5)
        # coordinate transform to global
        f_Eic = self.xf2F(f_Eic)
        _, n_i, n_c = f_Eic.shape
        f_Ei = f_Eic.reshape(-1, n_i * n_c)
        o_E = self.o_Eia.reshape(-1, n_i * n_c)
        return o_E.flatten(), f_Ei.flatten()


    def map_field_to_K(self, D_Est):
        # print('map_field_to_K')
        w_m = self.fets.w_m  # Gauss points weights
        B_Emiabo, det_J_Fm = self.B_Emiabo

        B_Empf = self._get_B_Empf()
        k2_Emop = self.integ_factor * np.einsum('m, Empf, Ept, Emtq, Em -> Emfq', w_m, B_Empf, D_Est, B_Empf,
                                                   det_J_Fm)
        k2_Eop = np.sum(k2_Emop, axis=1)

        # o_Ei = self.o_Eia.reshape(-1, 3 * 5)
        # print('o_Ei:', o_Ei)
        # # # o_Ei = self.o_Ia[self.mesh.I_Fi].reshape(-1, 3 * self.fets.n_nodal_dofs)
        # # o_Ei = self.o_Ia[self.F_N].reshape(-1, 3 * self.fets.n_nodal_dofs)
        #
        # return SysMtxArray(mtx_arr=k2_Eop, dof_map_arr=o_Ei)


        K_Eiejf = k2_Eop.reshape(-1, 3, 5, 3, 5)
        K_Eicjd = self.xk2K(K_Eiejf)
        _, n_i, n_c, n_j, n_d = K_Eicjd.shape
        K_Eij = K_Eicjd.reshape(-1, n_i * n_c, n_j * n_d)
        o_Ei = self.o_Eia.reshape(-1, n_i * n_c)
        return SysMtxArray(mtx_arr=K_Eij, dof_map_arr=o_Ei)


    # O_Eo = tr.Property(tr.Array, depends_on='X, L, F')
    # @tr.cached_property
    # def _get_O_Eo(self):
    #     return self.o_Ia[self.F_N].reshape(-1, 3 * self.fets.n_nodal_dofs)

    # =========================================================================
    # Property operators for initial configuration
    # =========================================================================
    F0_normals = tr.Property(tr.Array, depends_on='X, L, F')
    r'''Normal facet vectors.
    '''
    @tr.cached_property
    def _get_F0_normals(self):
        x_F = self.x_0[self.F]
        N_deta_ip = self.Na_deta
        r_deta = np.einsum('ajK,IKi->Iaij', N_deta_ip, x_F)
        Fa_normals = np.einsum('Iai,Iaj,ijk->Iak',
                               r_deta[..., 0], r_deta[..., 1], EPS)
        return np.sum(Fa_normals, axis=1)

    sign_normals = tr.Property(tr.Array, depends_on='X,L,F')
    r'''Orientation of the normal in the initial state.
    This array is used to switch the normal vectors of the faces
    to be oriented in the positive sense of the z-axis.
    '''
    @tr.cached_property
    def _get_sign_normals(self):
        return np.sign(self.F0_normals[:, 2])

    F_N = tr.Property(tr.Array, depends_on='X,L,F')
    r'''Counter-clockwise enumeration.
    '''
    @tr.cached_property
    def _get_F_N(self):
        # TODO: following test may be not valid in 3D shells case! other condition is to be taken
        turn_facets = np.where(self.sign_normals < 0)
        F_N = np.copy(self.F)
        F_N[turn_facets, :] = self.F[turn_facets, ::-1]
        return F_N

    F_normals = tr.Property(tr.Array, depends_on=INPUT)
    r'''Get the normals of the facets.
    '''
    @tr.cached_property
    def _get_F_normals(self):
        n = self.Fa_normals
        return np.sum(n, axis=1)

    F_normals_0 = tr.Property(tr.Array, depends_on=INPUT)
    r'''Get the normals of the facets.
    '''
    @tr.cached_property
    def _get_F_normals_0(self):
        n = self.Fa_normals_0
        return np.sum(n, axis=1)

    norm_F_normals = tr.Property(tr.Array, depends_on=INPUT)
    r'''Get the normed normals of the facets.
    '''
    @tr.cached_property
    def _get_norm_F_normals(self):
        n = self.F_normals
        mag_n = np.sqrt(np.einsum('...i,...i', n, n))
        return n / mag_n[:, np.newaxis]

    norm_F_normals_0 = tr.Property(tr.Array, depends_on=INPUT)
    r'''Get the normed normals of the facets.
    '''
    @tr.cached_property
    def _get_norm_F_normals_0(self):
        n = self.F_normals_0
        mag_n = np.sqrt(np.einsum('...i,...i', n, n))
        return n / mag_n[:, np.newaxis]

    F_normals_du = tr.Property(tr.Array, depends_on=INPUT)
    r'''Get the normals of the facets.
    '''
    @tr.cached_property
    def _get_F_normals_du(self):
        n_du = self.Fa_normals_du
        return np.sum(n_du, axis=1)

    F_area = tr.Property(tr.Array, depends_on=INPUT)
    r'''Get the surface area of the facets.
    '''
    @tr.cached_property
    def _get_F_area(self):
        a = self.Fa_area
        A = np.einsum('a,Ia->I', self.eta_w, a)
        return A

    # =========================================================================
    # Potential energy
    # =========================================================================
    F_V = tr.Property(tr.Array, depends_on=INPUT)
    r'''Get the total potential energy of gravity for each facet
    '''
    @tr.cached_property
    def _get_F_V(self):
        eta_w = self.eta_w
        a = self.Fa_area
        ra = self.Fa_r
        F_V = np.einsum('a,Ia,Ia->I', eta_w, ra[..., 2], a)
        return F_V

    F_V_du = tr.Property(tr.Array, depends_on=INPUT)
    r'''Get the derivative of total potential energy of gravity for each facet
    with respect to each node and displacement component [FIi]
    '''
    @tr.cached_property
    def _get_F_V_du(self):
        r = self.Fa_r
        a = self.Fa_area
        a_dx = self.Fa_area_du
        r3_a_dx = np.einsum('Ia,IaJj->IaJj', r[..., 2], a_dx)
        N_eta_ip = self.Na
        r3_dx = np.einsum('aK,KJ,j->aJj', N_eta_ip, DELTA, DELTA[2, :])
        a_r3_dx = np.einsum('Ia,aJj->IaJj', a, r3_dx)
        F_V_du = np.einsum('a,IaJj->IJj', self.eta_w, (a_r3_dx + r3_a_dx))
        return F_V_du

    # =========================================================================
    # Line vectors
    # =========================================================================

    F_L_vectors_0 = tr.Property(tr.Array, depends_on=INPUT)
    r'''Get the cycled line vectors around the facet
    The cycle is closed - the first and last vector are identical.

    .. math::
        v_{pld} \;\mathrm{where} \; p\in\mathcal{F}, l\in (0,1,2), d\in (0,1,2)

    with the indices :math:`p,l,d` representing the facet, line vector around
    the facet and and vector component, respectively.
    '''
    @tr.cached_property
    def _get_F_L_vectors_0(self):
        F_N = self.F_N  # F_N is cycled counter clockwise
        return self.x_0[F_N[:, (1, 2, 0)]] - self.x_0[F_N[:, (0, 1, 2)]]

    F_L_vectors = tr.Property(tr.Array, depends_on=INPUT)
    r'''Get the cycled line vectors around the facet
    The cycle is closed - the first and last vector are identical.


    .. math::
        v_{pld} \;\mathrm{where} \; p\in\mathcal{F}, l\in (0,1,2), d\in (0,1,2)

    with the indices :math:`p,l,d` representing the facet, line vector around
    the facet and and vector component, respectively.
    '''
    @tr.cached_property
    def _get_F_L_vectors(self):
        F_N = self.F_N  # F_N is cycled counter clockwise
        return self.x[F_N[:, (1, 2, 0)]] - self.x[F_N[:, (0, 1, 2)]]

    F_L_vectors_du = tr.Property(tr.Array, depends_on=INPUT)
    r'''Get the derivatives of the line vectors around the facets.

    .. math::
        \pard{v_{pld}}{x_{Ie}} \; \mathrm{where} \;
        p \in \mathcal{F},  \in (0,1,2), d\in (0,1,2), I\in \mathcal{N},
        e \in (0,1,3)

    with the indices :math:`p,l,d,I,e` representing the facet,
    line vector around the facet and and vector component,
    node vector and and its component index,
    respectively.

    This array works essentially as an index function delivering -1
    for the components of the first node in each dimension and +1
    for the components of the second node
    in each dimension.

    For a facet :math:`p` with lines :math:`l` and component :math:`d` return
    the derivatives with respect to the displacement of the node :math:`I`
    in the direction :math:`e`.

    .. math::
        \bm{a}_1 = \bm{x}_2 - \bm{x}_1 \\
        \bm{a}_2 = \bm{x}_3 - \bm{x}_2 \\
        \bm{a}_3 = \bm{x}_1 - \bm{x}_3

    The corresponding derivatives are then

    .. math::
        \pard{\bm{a}_1}{\bm{u}_1} = -1, \;\;\;
        \pard{\bm{a}_1}{\bm{u}_2} = 1 \\
        \pard{\bm{a}_2}{\bm{u}_2} = -1, \;\;\;
        \pard{\bm{a}_2}{\bm{u}_3} = 1 \\
        \pard{\bm{a}_3}{\bm{u}_3} = -1, \;\;\;
        \pard{\bm{a}_3}{\bm{u}_1} = 1 \\

    '''

    def _get_F_L_vectors_du(self):
        return self.L_vectors_du[self.F_L]

    F_L_vectors_dul = tr.Property(tr.Array, depends_on=INPUT)

    def _get_F_L_vectors_dul(self):
        return self.L_vectors_dul[self.F_L]

    norm_F_L_vectors = tr.Property(tr.Array, depends_on=INPUT)
    r'''Get the cycled line vectors around the facet
    The cycle is closed - the first and last vector are identical.
    '''
    @tr.cached_property
    def _get_norm_F_L_vectors(self):
        v = self.F_L_vectors
        mag_v = np.sqrt(np.einsum('...i,...i', v, v))
        return v / mag_v[..., np.newaxis]

    norm_F_L_vectors_du = tr.Property(tr.Array, depends_on=INPUT)
    '''Get the derivatives of cycled line vectors around the facet
    '''
    @tr.cached_property
    def _get_norm_F_L_vectors_du(self):
        v = self.F_L_vectors
        v_du = self.F_L_vectors_du  # @UnusedVariable
        mag_v = np.einsum('...i,...i', v, v)  # @UnusedVariable
        # @todo: finish the chain rule
        raise NotImplemented

    # =========================================================================
    # Orthonormal basis of each facet.
    # =========================================================================
    F_L_bases = tr.Property(tr.Array, depends_on=INPUT)
    r'''Line bases around a facet.
    '''
    @tr.cached_property
    def _get_F_L_bases(self):
        l = self.norm_F_L_vectors
        n = self.norm_F_normals
        lxn = np.einsum('...li,...j,...kij->...lk', l, n, EPS)
        n_ = n[:, np.newaxis, :] * np.ones((1, 3, 1), dtype='float_')
        T = np.concatenate([l[:, :, np.newaxis, :],
                            -lxn[:, :, np.newaxis, :],
                            n_[:, :, np.newaxis, :]], axis=2)
        return T

    F_L_bases_du = tr.Property(tr.Array, depends_on=INPUT)
    r'''Derivatives of the line bases around a facet.
    '''
    @tr.cached_property
    def _get_F_L_bases_du(self):
        '''Derivatives of line bases'''
        raise NotImplemented

    # =========================================================================
    # Sector angles
    # =========================================================================
    F_theta = tr.Property(tr.Array, depends_on=INPUT)
    '''Get the sector angles :math:`\theta`  within a facet.
    '''
    @tr.cached_property
    def _get_F_theta(self):
        v = self.F_L_vectors
        a = -v[:, (2, 0, 1), :]
        b = v[:, (0, 1, 2), :]
        return get_theta(a, b)

    F_theta_du = tr.Property(tr.Array, depends_on=INPUT)
    r'''Get the derivatives of sector angles :math:`\theta` within a facet.
    '''
    @tr.cached_property
    def _get_F_theta_du(self):
        v = self.F_L_vectors
        v_du = self.F_L_vectors_du

        a = -v[:, (2, 0, 1), :]
        b = v[:, (0, 1, 2), :]
        a_du = -v_du[:, (2, 0, 1), ...]
        b_du = v_du[:, (0, 1, 2), ...]

        return get_theta_du(a, a_du, b, b_du)

    Fa_normals_du = tr.Property
    '''Get the derivatives of the normals with respect
    to the node displacements.
    '''

    def _get_Fa_normals_du(self):
        x_F = self.x[self.F_N]
        N_deta_ip = self.Na_deta
        NN_delta_eps_x1 = np.einsum('aK,aL,KJ,dli,ILl->IaiJd',
                                    N_deta_ip[:, 0, :], N_deta_ip[:, 1, :],
                                    DELTA, EPS, x_F)
        NN_delta_eps_x2 = np.einsum('aK,aL,LJ,kdi,IKk->IaiJd',
                                    N_deta_ip[:, 0, :], N_deta_ip[:, 1, :],
                                    DELTA, EPS, x_F)
        n_du = NN_delta_eps_x1 + NN_delta_eps_x2
        return n_du

    Fa_area_du = tr.Property
    '''Get the derivatives of the facet area with respect
    to node displacements.
    '''

    def _get_Fa_area_du(self):
        a = self.Fa_area
        n = self.Fa_normals
        n_du = self.Fa_normals_du
        a_du = np.einsum('Ia,Iak,IakJd->IaJd', 1 / a, n, n_du)
        return a_du

    Fa_normals = tr.Property
    '''Get normals of the facets.
    '''

    def _get_Fa_normals(self):

        # TODO, check this change
        # x_F = self.x[self.mesh.I_Fi]
        x_F = self.x[self.F_N]

        N_deta_ip = self.Na_deta
        r_deta = np.einsum('ajK,IKi->Iaij', N_deta_ip, x_F)
        return np.einsum('Iai,Iaj,ijk->Iak',
                         r_deta[..., 0], r_deta[..., 1], EPS)

    Fa_normals_0 = tr.Property
    '''Get normals of the facets.
    '''

    def _get_Fa_normals_0(self):
        x_F = self.x_0[self.F_N]
        N_deta_ip = self.Na_deta
        r_deta = np.einsum('ajK,IKi->Iaij', N_deta_ip, x_F)
        return np.einsum('Iai,Iaj,ijk->Iak',
                         r_deta[..., 0], r_deta[..., 1], EPS)

    Fa_area = tr.Property
    '''Get the surface area of the facets.
    '''

    def _get_Fa_area(self):
        n = self.Fa_normals
        a = np.sqrt(np.einsum('Iai,Iai->Ia', n, n))
        return a

    Fa_r = tr.Property
    '''Get the reference vector to integrations point in each facet.
    '''

    def _get_Fa_r(self):
        x_F = self.x[self.F_N]
        N_eta_ip = self.Na
        r = np.einsum('aK,IKi->Iai', N_eta_ip, x_F)
        return r

    # =========================================================================
    # Interier as level set
    # =========================================================================




import traits.api as tr
from ibvpy.fets import FETSEval
from ibvpy.mathkit.tensor import DELTA23_ab
import numpy as np


class FETS2D3U1M(FETSEval):
    r'''Triangular, three-node element.
    '''

    vtk_r = tr.Array(np.float_, value=[[1, 0, 0], [0, 1, 0], [0, 0, 1]])
    vtk_cells = [[0, 1, 2]]
    vtk_cell_types = 'Triangle'
    vtk_cell = [0, 1, 2]
    vtk_cell_type = 'Triangle'

    vtk_expand_operator = tr.Array(np.float_, value=DELTA23_ab)

    # =========================================================================
    # Surface integrals using numerical integration
    # =========================================================================
    # i is gauss point index, p is point coords in natural coords (zeta_1, zeta_2, zeta_3)
    eta_ip = tr.Array('float_')
    r'''Integration points within a triangle.
    '''

    def _eta_ip_default(self):
        # here, just one integration point in the middle of the triangle (zeta_1 = 1/3, zeta_2 = 1/3, zeta_3 = 1/3)
        return np.array([[1. / 3., 1. / 3., 1. / 3.]], dtype='f')

    w_m = tr.Array('float_')
    r'''Weight factors for numerical integration.
    '''

    def _w_m_default(self):
        return np.array([1. / 2.], dtype='f')

    n_m = tr.Int(1)
    r'''Number of integration points.
    '''
    @tr.cached_property
    def _get_w_m(self):
        return len(self.w_m)

    n_nodal_dofs = tr.Int(3)

    # N_im = tr.Property(depends_on='eta_ip')
    # r'''Shape function values in integration points.
    # '''
    # @tr.cached_property
    # def _get_N_im(self):
    #     eta = self.eta_ip
    #     return np.array([eta[:, 0], eta[:, 1], 1 - eta[:, 0] - eta[:, 1]],
    #                     dtype='f')

    dN_imr = tr.Property(depends_on='eta_ip')
    r'''Derivatives of the shape functions in the integration points.
    '''
    @tr.cached_property
    def _get_dN_imr(self):
        dN_mri = np.array([[[1, 0, -1],         # dN1/d_zeta1, dN2/d_zeta1, dN3/d_zeta1
                          [0, 1, -1]],          # dN1/d_zeta2, dN2/d_zeta2, dN3/d_zeta2
                          ], dtype=np.float_)
        return np.einsum('mri->imr', dN_mri)

    dN_inr = tr.Property(depends_on='eta_ip')
    r'''Derivatives of the shape functions in the integration points.
    '''
    @tr.cached_property
    def _get_dN_inr(self):
        return self.dN_imr

    vtk_expand_operator = tr.Array(value=[1,1,0])
    vtk_node_cell_data = tr.Array
    vtk_ip_cell_data = tr.Array



import bmcs_utils.api as bu
import k3d
import numpy as np
import traits.api as tr
from bmcs_shell.folding.geometry.wb_shell_geometry import WBShellGeometry4P
from ibvpy.bcond import BCDof
import os

class BoundaryConditions(bu.Model):
    name = 'BoundaryConditions'
    plot_backend = 'k3d'

    id = bu.Str
    n_nodal_dofs = 5

    geo = bu.Instance(WBShellGeometry4P, ())

    bc_input = bu.Str(r'0, 0, 0,,')
    f_input = bu.Str(r',, -20000,,')
    add_bc_btn = bu.Button()
    add_bc_btn_editor = bu.ButtonEditor(icon='plus')
    add_force_btn = bu.Button()
    add_force_btn_editor = bu.ButtonEditor(icon='plus')
    save_geo_and_bc = bu.Button()
    active_button = tr.Str

    bc_node_3d_obj_map = {}
    force_node_3d_obj_map = {}
    pb = tr.Any(desc='Plot backend')

    bc_fixed_array = bu.Array(BC=True) # [[node_idx, bc_x, bc_y, bc_z...]]
    bc_loaded_array = bu.Array(BC=True) # [[node_idx, f_x, f_y, f_z...]]

    draw_points_pending_after_cache_load = bu.Bool

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        if os.path.isfile(self._get_bc_fixed_array_path(ext=True)):
            self.bc_loaded_array = np.load(self._get_bc_loaded_array_path(ext=True))
            self.bc_fixed_array = np.load(self._get_bc_fixed_array_path(ext=True))
            self.draw_points_pending_after_cache_load = True
            print('BCs were loaded from files: ' + self._get_bcs_id_cache_dir() + '_...')

    @tr.observe('add_bc_btn')
    def add_bc_click(self, event=None):
        self._switch_btn('add_bc_btn', self.add_bc_btn_editor, self.add_force_btn_editor)

    @tr.observe('add_force_btn')
    def add_force_click(self, event=None):
        self._switch_btn('add_force_btn', self.add_force_btn_editor, self.add_bc_btn_editor)

    @tr.observe('save_geo_and_bc')
    def save_geo_and_bc_click(self, event=None):
        np.save(self._get_bc_fixed_array_path(ext=False), self.bc_fixed_array)
        np.save(self._get_bc_loaded_array_path(ext=False), self.bc_loaded_array)
        print('BCs were saved successfully to files: ' + self._get_bcs_id_cache_dir() + '_...')

    def _get_bc_fixed_array_path(self, ext=False):
        return self._get_bcs_id_cache_dir() + '_bc_fixed_array' + ('.npy' if ext else '')

    def _get_bc_loaded_array_path(self, ext=False):
        return self._get_bcs_id_cache_dir() + '_bc_loaded_array' + ('.npy' if ext else '')

    def _get_bcs_id_cache_dir(self):
        cache_dir = bu.data_cache.dir
        bcs_cache_dir = os.path.join(cache_dir, 'bcs')
        if not os.path.exists(bcs_cache_dir):
            os.makedirs(bcs_cache_dir)
        return os.path.join(bcs_cache_dir, self.id)

    def _switch_btn(self, button_str, btn_editor, other_editor):
        if self.active_button == button_str:
            self.active_button = ''
            self.pb.plot_fig.mode = 'view'
            self.pb.plot_fig.camera_auto_fit = True
            btn_editor.widget.button_style = ''
        else:
            self.active_button = button_str
            self.pb.plot_fig.mode = 'callback'
            self.pb.plot_fig.camera_auto_fit = False
            btn_editor.widget.button_style = 'success'
            other_editor.widget.button_style = ''

    def _closest_node(self, node, nodes):
        nodes = np.asarray(nodes)
        dist_2 = np.sum((nodes - node) ** 2, axis=1)
        return np.argmin(dist_2)

    ipw_view = bu.View(
        bu.Item('bc_input', latex=r'\mathrm{BC~values}'),
        bu.Item('add_bc_btn', editor=add_bc_btn_editor),
        bu.Item('f_input', latex=r'\mathrm{F~values}'),
        bu.Item('add_force_btn', editor=add_force_btn_editor),
        bu.Item('save_geo_and_bc', editor=bu.ButtonEditor(icon='download')),
    )

    def setup_plot(self, pb):
        self.pb = pb
        self.geo.setup_plot(pb)

        vertices = self.geo.X_Ia

        # Each time one clicks on the tap in the tree, setup_plot() gets called
        # Here, plot is empty, update 3d points (they could've changed in geo)
        for node_idx, _ in self.bc_node_3d_obj_map.items():
            # Following will override the corresponding 3d_point in the map
            self._add_bc_3d_point(node_idx, vertices[node_idx])
        for node_idx, _ in self.force_node_3d_obj_map.items():
            # Following will override the corresponding 3d_point in the map
            self._add_force_3d_point(node_idx, vertices[node_idx])

        if self.draw_points_pending_after_cache_load:
            for loaded_bc in self.bc_loaded_array:
                idx = int(loaded_bc[0])
                self._add_force_3d_point(idx, vertices[idx])
            for fixed_bc in self.bc_fixed_array:
                idx = int(fixed_bc[0])
                self._add_bc_3d_point(idx, vertices[idx])
            self.draw_points_pending_after_cache_load = False

        def mesh_click(params):
            """ (example) params = {'msg_type': 'hover_callback',
                                    'position': [x, y, z],
                                    'normal': [0, 0, 1],
                                    'distance': 20,
                                    'face_index': 0,
                                    'face': [0, 1, 2]} """
            vertices = self.geo.X_Ia
            touch_pos = np.array(params['position'])
            idx = self._closest_node(touch_pos, vertices)
            point = vertices[idx]

            if self.active_button == 'add_bc_btn':
                if idx in self.bc_node_3d_obj_map:
                    self.pb.plot_fig -= self.bc_node_3d_obj_map[idx]
                    self.bc_node_3d_obj_map.pop(idx)
                    self._remove_bc(idx)
                else:
                    if self._add_bc(idx):
                        self._add_bc_3d_point(idx, point)
            elif self.active_button == 'add_force_btn':
                if idx in self.force_node_3d_obj_map:
                    self.pb.plot_fig -= self.force_node_3d_obj_map[idx]
                    self.force_node_3d_obj_map.pop(idx)
                    self._remove_force(idx)
                else:
                    if self._add_force(idx):
                        self._add_force_3d_point(idx, point)

        pb.objects['k3d_mesh'].click_callback = mesh_click
        # wb_mesh_1.hover_callback = foo

    def update_plot(self, pb):
        self.geo.update_plot(pb)

    def _add_bc_3d_point(self, idx, point):
        k3d_point = k3d.points(point, point_size=100, color=0x3b3b3b)
        self.pb.plot_fig += k3d_point
        self.bc_node_3d_obj_map[idx] = k3d_point

    def _add_force_3d_point(self, idx, point):
        k3d_point = k3d.points(point, point_size=100, color=0xFF0000)
        self.pb.plot_fig += k3d_point
        self.force_node_3d_obj_map[idx] = k3d_point

    def _parse_text_to_xyz(self, text):
        try:
            dofs_values = np.empty(self.n_nodal_dofs) * np.nan

            text = text.replace(' ', '')
            commas_idxs = [i for i, char_ in enumerate(text) if char_ == ',']

            if len(commas_idxs) != self.n_nodal_dofs - 1:
                raise ValueError('Num of commas in BC input box should be (dofs per node -1)!')

            start_idx = 0
            for i, commas_idx in enumerate(commas_idxs):
                dof_value_txt = text[start_idx:commas_idx]
                if dof_value_txt:
                    dofs_values[i] = float(dof_value_txt)
                start_idx = commas_idx + 1
                if i == len(commas_idxs) - 1:
                    dof_value_txt = text[start_idx:]
                    if dof_value_txt:
                        dofs_values[i + 1] = float(dof_value_txt)
            return dofs_values
        except:
            raise ValueError('An invalid value has been provided in bc text box!')

    def _remove_force(self, idx):
        # Remove the row corresponding to idx
        self.bc_loaded_array = self.bc_loaded_array[self.bc_loaded_array[:, 0] != idx]

    def _remove_bc(self, idx):
        # Remove the row corresponding to idx
        self.bc_fixed_array = self.bc_fixed_array[self.bc_fixed_array[:, 0] != idx]

    def _add_bc(self, idx):
        dofs_values = self._parse_text_to_xyz(self.bc_input)
        dofs_nans = np.empty_like(dofs_values) * np.nan
        if (dofs_values == dofs_nans).all():
            # bc should not be added, empty field!
            return False
        if self.bc_fixed_array.size == 0:
            self.bc_fixed_array = np.array([[idx, *dofs_values]])
        else:
            self.bc_fixed_array = np.append(self.bc_fixed_array, [[idx, *dofs_values]], axis=0)
        return True

    def _add_force(self, idx):
        dofs_values = self._parse_text_to_xyz(self.f_input)
        dofs_nans = np.empty_like(dofs_values) * np.nan
        if (dofs_values == dofs_nans).all():
            # bc should not be added, empty field!
            return False
        if self.bc_loaded_array.size == 0:
            self.bc_loaded_array = np.array([[idx, *dofs_values]])
        else:
            self.bc_loaded_array = np.append(self.bc_loaded_array, [[idx, *dofs_values]], axis=0)
        return True

    bc_loaded_method = bu.Str('manual')

    bc_loaded = tr.Property(depends_on="state_changed")
    # @tr.cached_property
    def _get_bc_loaded(self):
        # This whole method can be moved to BoundaryConditions
        if self.bc_loaded_method == 'automatic':
            return self._get_bc_loaded_automatic()
        elif self.bc_loaded_method == 'manual':
            return self._get_dofs(self.bc_loaded_array, 'f')

    def _get_bc_loaded_automatic(self):
        F = bu.Float(-1000, BC=True)
        xdomain = self.xdomain
        ix2 = int((self.n_phi_plus) / 2)
        F_I = xdomain.mesh.I_CDij[ix2, :, 0, :].flatten()
        _, idx_remap = xdomain.mesh.unique_node_map
        loaded_nodes = idx_remap[F_I]  # loaded_nodes = xdomain.bc_J_F
        loaded_dofs = (loaded_nodes[:, np.newaxis] * 3 + 2).flatten()
        bc_loaded = [BCDof(var='f', dof=dof, value=F)
                     for dof in loaded_dofs]
        return bc_loaded, loaded_nodes, loaded_dofs

    bc_fixed = tr.Property(depends_on="state_changed")
    # @tr.cached_property
    def _get_bc_fixed(self):
        if self.bc_fixed_method == 'automatic':
            return self._get_bc_fixed_automatic()
        elif self.bc_fixed_method == 'manual':
            return self._get_dofs(self.bc_fixed_array, 'u')

    def _get_dofs(self, bc_or_f_array, type):
        if bc_or_f_array.size == 0:
            return [], np.array([]), np.array([])

        """ Note: Naming is for bc but it works also for forces """
        # Note: bcs.bc_fixed gives nodes indices in geo which are the same in mesh so no mapping is needed!
        bcs_N_ = np.copy(bc_or_f_array)  # [[node_idx, bc_x, bc_y, bc_z], ...]
        fixed_nodes_N = np.copy(bcs_N_[:, 0])

        unfiltered_dofs_Nd = bcs_N_[:, 0]
        unfiltered_dofs_Nd = unfiltered_dofs_Nd[:, np.newaxis] * self.n_nodal_dofs + np.arange(self.n_nodal_dofs)[np.newaxis, :]
        np.zeros_like(bcs_N_[:, 1:])
        unfiltered_dofs_N_ = np.zeros_like(bcs_N_)
        unfiltered_dofs_N_[:, 1:] = unfiltered_dofs_Nd

        # Replacing nodes indices with nan so they can be eliminated too when extracting dofs
        bcs_N_[:, 0] = np.full_like(bcs_N_[:, 0], np.nan)

        no_nan_mask = ~np.isnan(bcs_N_)
        dofs = unfiltered_dofs_N_[no_nan_mask].astype(np.uint32)
        dofs_values = bcs_N_[no_nan_mask]

        bc_fixed = [BCDof(var=type, dof=dof, value=val)
                    for dof, val in zip(dofs, dofs_values)]

        return bc_fixed, fixed_nodes_N, dofs

    bc_fixed_method = bu.Str('manual')

    def _get_bc_fixed_automatic(self):
        xdomain = self.xdomain
        # Node indicies for nodes that are fixed in x, y, z directions
        fixed_xyz_nodes = xdomain.bc_J_xyz
        # Node indicies for nodes that are fixed in x direction
        fixed_x_nodes = xdomain.bc_J_x
        fixed_nodes = np.unique(np.hstack([fixed_xyz_nodes, fixed_x_nodes]))
        fixed_xyz_dofs = (fixed_xyz_nodes[:, np.newaxis] * 3 + np.arange(3)[np.newaxis, :]).flatten()
        fixed_x_dofs = (fixed_x_nodes[:, np.newaxis] * 3).flatten()
        fixed_dofs = np.unique(np.hstack([fixed_xyz_dofs, fixed_x_dofs]))
        bc_fixed = [BCDof(var='u', dof=dof, value=0)
                    for dof in fixed_dofs]
        return bc_fixed, fixed_nodes, fixed_dofs


'''
Created on Aug 28, 2014

@author: rch
'''

import numpy as np


def get_cos_theta(a, b):
    r'''Get the cosine of an angle between two vectors :math:`a` and :math:`b`

    .. math::
        \gamma_c =
        \cos{(\theta)}
        =
        \frac{ \dotp{a}{b} }
            { \norm{a} \norm{b} }
        =
        \frac{a_d b_d}
            { \sqrt{a_d a_d} \sqrt{b_d b_d} }
        :label: gamma

    It is convenient to implement this function in a vectorized
    form so that two arrays of vectors can be supplied and
    evaluated using several sum-product evaluations over their
    last index :math:`d` standing for the spatial dimension of
    the vectors.

    Such an evaluation can be conveniently done using the
    method
    `numpy.einsum(subscript_mappings, *operands)
    <http://docs.scipy.org/doc/numpy/reference/generated/numpy.einsum.html>`_.
    Given two compatible arrays (n-dimensional) :math:`a`
    and :math:`b` of vectors with last index :math:`d`
    labeling their spatial component the array :math:`\gamma_c`
    of angle cosines between each pair of vectors in :math:`a`, :math:`b`.
    gets calculated as::

        ab = np.einsum('...d,...d->...', a, b)
        sqrt_aa = np.sqrt(np.einsum('...d,...d->...', a, a))
        sqrt_bb = np.sqrt(np.einsum('...d,...d->...', b, b))
        sqrt_aa_x_sqrt_bb = sqrt_aa * sqrt_bb
        gamma = ab / sqrt_aa_x_sqrt_bb)

    '''
    ab = np.einsum('...d,...d->...', a, b)
    sqrt_aa = np.sqrt(np.einsum('...d,...d->...', a, a))
    sqrt_bb = np.sqrt(np.einsum('...d,...d->...', b, b))
    c_ab = ab / (sqrt_aa * sqrt_bb)
    return c_ab


def get_cos_theta_du(a, a_du, b, b_du):
    r'''Get the derivatives of directional cosines
    between two vectors :math:`a` and :math:`b`.

    Given two arrays (n-dimensional) of vectors :math:`a, b` and their
    derivatives a_du and b_du with respect to the nodal displacments du
    return the derivatives of the cosine between mutual angles theta between
    a and b with respect to the node displacement du.

    .. math::
        \pard{ \gamma_c }
            {u_{Ie}}
        =
        \frac{1}
            { \left( \norm{a} \norm{b} \right)^2 }
        \left[
            \pard{ \left( a_d  b_d \right)}
                {u_{Ie}}
            \left( \norm{a} \norm{b} \right)
            -
            \pard{ \left( \norm{a} \norm{b} \right) }
                {u_{Ie}} \;
            \left( a_d b_d \right)
        \right]

    By factoring out the term
    :math:`\left( \norm{ a } \norm{ b } \right)`
    the expression reduces to a form

    .. math::
        \pard{ \gamma_c }
            {u_{Ie}}
        =
        \frac{1}
            { \norm{a}\norm{b}  }
        \left[
            \pard{ \left(a_d  b_d \right)}
                {u_{Ie}}
            -
            \pard{ \left( \norm{a}\norm{b} \right) }
                {u_{Ie}}\;
            \gamma_c
        \right]
        :label: gamma_du

    The terms with derivatives can be further elaborated to

    .. math::
        \pard{ \left(a_d  b_d \right)}
            {u_{Ie}}
        =
        \pard{a_d }
            {u_{Ie}} \;
        b_d
        +
        \pard{b_d }{u_{Ie}} \;
        a_d
        \;\;\mathrm{and}\;\;
        \pard{\left( \norm{a}\norm{b}\right) }
            {u_{Ie}}
        =
        \pard{ \norm{a} }
            {u_{Ie}} \;
        \norm{b}
        +
        \pard{ \norm{b}  }
            {u_{Ie}}\;
        \norm{a}

    The derivative of a vector norm in a summation form can be rewritten as

    .. math::
        \pard{\norm{a}}
            {u_{Ie}}
        =
        \pard{ \sqrt{ a_d a_d } }
            {u_{Ie}}
        =
        \frac{1}
            {\sqrt{ a_d a_d } }
        \left(
            \pard{a_{d}}
                {u_{Ie}}
            a_{d}
        \right)

    The corresponding code for a vectoried evaluation for the supplied
    multidimensional arrays of
    ``a, a_du, b, b_du`` looks as follows::

        ab_du = np.einsum('...dIe,...d->...Ie', a_du, b) + np.einsum('...dIe,...d->...Ie', b_du, a)
        sqrt_aa_du = 1 / sqrt_aa * np.einsum('...dIe,...d->...Ie', a_du, a)
        sqrt_bb_du = 1 / sqrt_bb * np.einsum('...dIe,...d->...Ie', b_du, b)
        sqrt_aa_x_sqrt_bb_du = (np.einsum(...Ie,...->...Ie', sqrt_aa_du, sqrt_bb) +
                                np.einsum(...Ie,...->...Ie', sqrt_bb_du, sqrt_aa)
        gamma_du = 1 / sqrt_aa_x_sqrt_bb * (ab_du - sqrt_aa_x_sqrt_bb_du * gamma)

    '''
    ab = np.einsum('...d,...d->...', a, b)
    norm_a = np.sqrt(np.einsum('...d,...d->...', a, a))
    norm_b = np.sqrt(np.einsum('...d,...d->...', b, b))
    norm_a_x_norm_b = norm_a * norm_b
    gamma = ab / norm_a_x_norm_b

    ab_du = (np.einsum('...eId,...d->...Ie', a_du, b) +
             np.einsum('...eId,...d->...Ie', b_du, a))
    gamma_norm_a_du_x_norm_b = np.einsum('...,...eId,...d->...Ie',
                                         gamma * norm_b / norm_a, a_du, a)
    gamma_norm_b_du_x_norm_a = np.einsum('...,...eId,...d->...Ie',
                                         gamma * norm_a / norm_b, b_du, b)
    gamma_norm_a_x_norm_b_du = (gamma_norm_a_du_x_norm_b +
                                gamma_norm_b_du_x_norm_a)
    gamma_du = np.einsum('...,...Ie->...Ie',
                         1. / norm_a_x_norm_b,
                         (ab_du - gamma_norm_a_x_norm_b_du))

    return gamma_du


def get_cos_theta_du2(a, a_du, b, b_du):
    r'''Given two arrays (n-dimensional) of vectors a, b and their
    derivatives a_du and b_du with respect to the nodal displacments du
    return the derivatives of the mutual angles theta between
    a and b with respect to the node displacement du.
    '''
    ab = np.einsum('...i,...i->...', a, b)
    sqrt_aa = np.sqrt(np.einsum('...i,...i->...', a, a))
    sqrt_bb = np.sqrt(np.einsum('...i,...i->...', b, b))
    sqrt_aa_x_sqrt_bb = sqrt_aa * sqrt_bb
    gamma = ab / sqrt_aa_x_sqrt_bb
    ab_du = (np.einsum('...iKj,...i->...Kj', a_du, b) +
             np.einsum('...iKj,...i->...Kj', b_du, a))

    gamma_bb__aa = gamma * sqrt_bb / sqrt_aa
    gamma_aa__bb = gamma * sqrt_aa / sqrt_bb
    gamma_aa_bb_du = (np.einsum('...,...iKj,...i->...Kj',
                                gamma_bb__aa, a_du, a) +
                      np.einsum('...,...iKj,...i->...Kj',
                                gamma_aa__bb, b_du, b))

    gamma_du = np.einsum(
        '...,...Kj->...Kj', 1. / sqrt_aa_x_sqrt_bb, (ab_du - gamma_aa_bb_du))
    return gamma_du


def get_theta(a, b):
    r'''
    Get the angle between two vectors :math:`a` and :math:`b`.
    Using the function ``get_cos_theta(a,b)`` delivering the cosine of the angle
    this method just evaluates the expression

    .. math::
        \theta = \arccos{( \gamma_c )}
        :label: theta

    realized by calling the vectorized numpy function::

        theta = np.arccos(gamma_c)

    '''
    gamma_c = get_cos_theta(a, b)
    theta = np.arccos(gamma_c)
    return theta


def get_theta_du(a, a_du, b, b_du):
    r'''
    Get the derivative of the angle between two vectors :math:`a` and :math:`b`
    with respect to ``du`` using the supplied chain derivatives ``a_du, b_du``.
    Using the function ``get_cos_theta(a,b)`` delivering the cosine of the angle
    this method evaluates the expression

    .. math::
        \frac{\partial \theta}{\partial u_{Ie}}
        =
        \frac{\partial \theta}{\partial \gamma_c}
        \cdot
        \frac{\partial \gamma_c}{\partial u_{Ie}}
        :label: theta_du

    where

    .. math::

        \frac{\partial \theta}{\partial \gamma_c}
        =
        - \frac{1}{ \sqrt{ 1 - \gamma_c^2}}

    ::

        theta_du = - 1 / np.sqrt(1 - gamma * gamma) * gamma_du
    '''
    gamma = get_cos_theta(a, b)
    gamma_du = get_cos_theta_du(a, a_du, b, b_du)
    theta_du = np.einsum(
        '...,...Ie->...Ie', -1. / np.sqrt(1. - gamma ** 2), gamma_du)
    return theta_du



import bmcs_utils.api as bu
import k3d
import traits.api as tr
import numpy as np
from bmcs_shell.folding.geometry.wb_cell.wb_cell import WBCell
from numpy import sin, cos, sqrt


class WBCell5ParamV2(WBCell):
    name = 'waterbomb cell 5p v2'

    plot_backend = 'k3d'

    gamma = bu.Float(np.pi / 6, GEO=True)
    beta = bu.Float(np.pi / 3, GEO=True)
    a = bu.Float(500, GEO=True)
    b = bu.Float(750, GEO=True)
    c = bu.Float(400, GEO=True)

    continuous_update = True

    ipw_view = bu.View(
        bu.Item('gamma', latex=r'\gamma', editor=bu.FloatRangeEditor(
            low=1e-6, high=np.pi / 2, n_steps=100, continuous_update=continuous_update)),
        bu.Item('beta', latex=r'\beta', editor=bu.FloatRangeEditor(
            low=1e-6, high=np.pi - 1e-6, n_steps=100, continuous_update=continuous_update)),
        bu.Item('a', latex='a', editor=bu.FloatRangeEditor(
            low=1e-6, high=2000, n_steps=100, continuous_update=continuous_update)),
        bu.Item('b', latex='b', editor=bu.FloatRangeEditor(
            low=1e-6, high=2000, n_steps=100, continuous_update=continuous_update)),
        bu.Item('c', latex='c', editor=bu.FloatRangeEditor(
            low=1e-6, high=2000, n_steps=100, continuous_update=continuous_update)),
        *WBCell.ipw_view.content,
    )

    X_Ia = tr.Property(depends_on='+GEO')
    '''Array with nodal coordinates I - node, a - dimension
    '''

    @tr.cached_property
    def _get_X_Ia(self):
        return self.get_cell_vertices()

    def get_cell_vertices(self, a=0.5, b=0.75, c=0.4, gamma=np.pi / 6, beta=np.pi / 3):
        a = self.a
        b = self.b
        c = self.c
        gamma = self.gamma
        beta = self.beta

        # phi1 is angle between OU_ur line and z axis
        cos_psi1 = ((b ** 2 - a ** 2) - a * sqrt(a ** 2 + b ** 2) * cos(beta)) / (b * sqrt(a ** 2 + b ** 2) * sin(beta))
        sin_psi1 = sqrt(
            a ** 2 * (3 * b ** 2 - a ** 2) + 2 * a * (b ** 2 - a ** 2) * sqrt(a ** 2 + b ** 2) * cos(beta) - (
                    a ** 2 + b ** 2) ** 2 * cos(beta) ** 2) / (b * sqrt(a ** 2 + b ** 2) * sin(beta))
        cos_psi5 = (sqrt(a ** 2 + b ** 2) * cos(beta) - a * cos(2 * gamma)) / (b * sin(2 * gamma))
        sin_psi5 = sqrt(b ** 2 + 2 * a * sqrt(a ** 2 + b ** 2) * cos(beta) * cos(2 * gamma) - (a ** 2 + b ** 2) * (
                cos(beta) ** 2 + cos(2 * gamma) ** 2)) / (b * sin(2 * gamma))
        cos_psi6 = (a - sqrt(a ** 2 + b ** 2) * cos(beta) * cos(2 * gamma)) / (
                sqrt(a ** 2 + b ** 2) * sin(beta) * sin(2 * gamma))
        sin_psi6 = sqrt(b ** 2 + 2 * a * sqrt(a ** 2 + b ** 2) * cos(beta) * cos(2 * gamma) - (a ** 2 + b ** 2) * (
                cos(beta) ** 2 + cos(2 * gamma) ** 2)) / (sqrt(a ** 2 + b ** 2) * sin(beta) * sin(2 * gamma))
        cos_psi1plus6 = cos_psi1 * cos_psi6 - sin_psi1 * sin_psi6
        sin_psi1plus6 = sin_psi1 * cos_psi6 + cos_psi1 * sin_psi6

        cos_phi1 = cos_psi1plus6
        cos_phi2 = cos_psi5
        cos_phi3 = cos_psi5
        cos_phi4 = cos_psi1plus6
        sin_phi1 = sin_psi1plus6
        sin_phi2 = sin_psi5
        sin_phi3 = sin_psi5
        sin_phi4 = sin_psi1plus6

        U_ur = np.array(
            [a * sin(gamma) - b * cos_phi1 * cos(gamma), b * sin_phi1, a * cos(gamma) + b * cos_phi1 * sin(gamma)])
        U_ul = np.array(
            [-a * sin(gamma) + b * cos_phi2 * cos(gamma), b * sin_phi2, a * cos(gamma) + b * cos_phi2 * sin(gamma)])
        U_lr = np.array(
            [a * sin(gamma) - b * cos_phi3 * cos(gamma), -b * sin_phi3, a * cos(gamma) + b * cos_phi3 * sin(gamma)])
        U_ll = np.array(
            [-a * sin(gamma) + b * cos_phi4 * cos(gamma), -b * sin_phi4, a * cos(gamma) + b * cos_phi4 * sin(gamma)])
        V_r = np.array([c * sin(gamma), 0, c * cos(gamma)])
        V_l = np.array([-c * sin(gamma), 0, c * cos(gamma)])

        X_Ia = np.vstack((np.zeros(3), U_lr, U_ll, U_ur, U_ul, V_r, V_l)).astype(np.float32)
        return X_Ia



import traits.api as tr
from ibvpy.fets import FETSEval
from ibvpy.mathkit.tensor import DELTA23_ab
import numpy as np


class FETS2DMITC(FETSEval):
    r'''MITC3 shell finite element:
        See https://www.sesamx.io/blog/standard_linear_triangular_shell_element/
        See http://dx.doi.org/10.1016/j.compstruc.2014.02.005
    '''

    vtk_r = tr.Array(np.float_, value=[[1, 0, 0], [0, 1, 0], [0, 0, 1]])
    vtk_cells = [[0, 1, 2]]
    vtk_cell_types = 'Triangle'
    vtk_cell = [0, 1, 2]
    vtk_cell_type = 'Triangle'

    vtk_expand_operator = tr.Array(np.float_, value=DELTA23_ab)

    # =========================================================================
    # Surface integrals using numerical integration
    # =========================================================================
    # i is point indices, p is point coords in natural coords (zeta_1, zeta_2, zeta_3)
    eta_ip = tr.Array('float_')
    r'''Integration points within a triangle.
    '''

    def _eta_ip_default(self):
        # We apply the 7-point Gauss integration to integrate exactly on the plane defined by rr and ss
        # [r, s, t] at each Gauss point (r, s, t) are natural coords in the shell element r,s in plane
        # and t along thickness
        # 2 Gauss points along thickness t
        # 7 Gauss points on the plane of the element
        return np.array([[1. / 3., 1. / 3., 0]], dtype='f')

    w_m = tr.Array('float_')
    r'''Weight factors for numerical integration.
    '''

    def _w_m_default(self):
        return np.array([1], dtype='f')

    # TODO, different node thickness in each node according to the original
    #  implementation can be easily integrated, but here the same thickness
    #  is used for simplicity.
    a = tr.Float(1.0,
                 label='thickness')

    n_m = tr.Property(depends_on='w_m')
    r'''Number of integration points.
    '''

    @tr.cached_property
    def _get_n_m(self):
        return len(self.w_m)

    n_nodal_dofs = tr.Int(5)

    dh_imr = tr.Property(depends_on='eta_ip')
    r'''Derivatives of the shape functions in the integration points.'''

    @tr.cached_property
    def _get_dh_imr(self):
        # Same for all Gauss points
        dh_ri = np.array([[-1, 1, 0],       # dh1/d_r, dh2/d_r, dh3/d_r
                           [-1, 0, 1],      # dh1/d_s, dh2/d_s, dh3/d_s
                           [0, 0, 0]],      # dh1/d_t, dh2/d_t, dh3/d_t
                          dtype=np.float_)
        dh_mri = np.tile(dh_ri, (self.n_m, 1, 1))

        return np.einsum('mri->imr', dh_mri)

    dht_imr = tr.Property(depends_on='eta_ip')
    r'''Derivatives of the (shape functions * t) in the integration points.
    '''

    @tr.cached_property
    def _get_dht_imr(self):
        # m: gauss points, r: r, s, t, and i: h1, h2, h3
        eta_ip = self.eta_ip
        dh_mri = np.array([[[-t, t, 0],         # (t*dh1)/d_r, (t*dh2)/d_r, (t*dh3)/d_r
                            [-t, 0, t],         # (t*dh1)/d_s, (t*dh2)/d_s, (t*dh3)/d_s
                            [1 - r - s, r, s]]  # (t*dh1)/d_t, (t*dh2)/d_t, (t*dh3)/d_t
                           for r, s, t in zip(eta_ip[:, 0], eta_ip[:, 1], eta_ip[:, 2])],
                          dtype=np.float_)
        return np.einsum('mri->imr', dh_mri)

    # dh_inr = tr.Property(depends_on='eta_ip')
    # r'''Derivatives of the shape functions in the integration points.
    # '''
    #
    # @tr.cached_property
    # def _get_dh_inr(self):
    #     return self.dh_imr
    #
    # def get_B(self):
    #     pass
    #
    # vtk_expand_operator = tr.Array(value=[1, 1, 0])
    # vtk_node_cell_data = tr.Array
    # vtk_ip_cell_data = tr.Array



"""

"""
import bmcs_utils.api as bu
import k3d
from bmcs_shell.folding.geometry.wb_cell.wb_cell_4p import \
    WBCell4Param, axis_angle_to_q, qv_mult
import traits.api as tr
import numpy as np
import json
import time


class WBTessellation4P(bu.Model):
    name = 'WB Tessellation 4P'

    wb_cell = bu.Instance(WBCell4Param, ())

    tree = ['wb_cell']

    plot_backend = 'k3d'

    n_phi_plus = bu.Int(5, GEO=True)
    n_x_plus = bu.Int(3, GEO=True)
    gamma = bu.Float(1.25, GEO=True)
    a = bu.Float(1000, GEO=True)
    a_high = bu.Float(2000)
    b = bu.Float(1000, GEO=True)
    b_high = bu.Float(2000)
    c = bu.Float(1000, GEO=True)
    c_high = bu.Float(2000)

    show_wireframe = bu.Bool(True, GEO=True)
    show_nodes = bu.Bool(False, GEO=True)
    show_node_labels = bu.Bool(False, GEO=True)
    WIREFRAME = 'k3d_mesh_wireframe'
    NODES = 'k3d_nodes'
    NODES_LABELS = 'k3d_nodes_labels'

    @tr.observe('+GEO', post_init=True)
    def update_wb_cell(self, event):
        self.wb_cell.trait_set(
            gamma=self.gamma,
            a=self.a,
            a_high=self.a_high,
            b=self.b,
            b_high=self.b_high,
            c=self.c,
            c_high=self.c_high,
        )

    ipw_view = bu.View(
        # bu.Item('wb_cell'),
        *WBCell4Param.ipw_view.content,
        bu.Item('n_phi_plus', latex = r'n_\phi'),
        bu.Item('n_x_plus', latex = r'n_x'),
        # bu.Item('show_wireframe'),
        # bu.Item('show_node_labels'),
        bu.Item('show_nodes'),
    )

    def get_phi_range(self, delta_phi):
        return np.arange(-(self.n_phi_plus - 1), self.n_phi_plus) * delta_phi

    def get_X_phi_range(self,delta_phi, R_0):
        """Given an array of angles and radius return an array of coordinates
        """
        phi_range = self.get_phi_range((delta_phi))
        return np.array([np.fabs(R_0) * np.sin(phi_range),
                         np.fabs(R_0) * np.cos(phi_range) + R_0]).T

    def get_X_x_range(self,delta_x):
        return np.arange(-(self.n_x_plus - 1), self.n_x_plus) * delta_x

    cell_map = tr.Property
    def _get_cell_map(self):
        delta_x = self.wb_cell.delta_x
        delta_phi = self.wb_cell.delta_phi
        R_0 = self.wb_cell.R_0

        X_x_range = self.get_X_x_range(delta_x)
        X_phi_range = self.get_X_phi_range(delta_phi, R_0)
        n_idx_x = len(X_x_range)
        n_idx_phi = len(X_phi_range)
        idx_x = np.arange(n_idx_x)
        idx_phi = np.arange(n_idx_phi)

        idx_x_ic = idx_x[(n_idx_x) % 2::2]
        idx_x_id = idx_x[(n_idx_x + 1) % 2::2]
        idx_phi_ic = idx_phi[(n_idx_phi) % 2::2]
        idx_phi_id = idx_phi[(n_idx_phi + 1) % 2::2]

        n_ic = len(idx_x_ic) * len(idx_phi_ic)
        n_id = len(idx_x_id) * len(idx_phi_id)

        n_cells = n_ic + n_id
        return n_cells, n_ic, n_id, idx_x_ic, idx_x_id, idx_phi_ic, idx_phi_id

    n_cells = tr.Property
    def _get_n_cells(self):
        n_cells, _, _, _, _, _, _ = self.cell_map
        return n_cells

    X_cells_Ia = tr.Property(depends_on='+GEO')
    '''Array with nodal coordinates of uncoupled cells
    I - node, a - dimension
    '''
    @tr.cached_property
    def _get_X_cells_Ia(self):

        delta_x = self.wb_cell.delta_x
        delta_phi = self.wb_cell.delta_phi
        R_0 = self.wb_cell.R_0

        X_Ia_wb_rot = np.copy(self.wb_cell.X_Ia)
        X_Ia_wb_rot[...,2] -= R_0
        X_cIa = np.array([X_Ia_wb_rot], dtype=np.float_)
        rotation_axes = np.array([[1, 0, 0]], dtype=np.float_)
        rotation_angles = self.get_phi_range(delta_phi)
        q = axis_angle_to_q(rotation_axes, rotation_angles)
        X_dIa = qv_mult(q, X_cIa)
        X_dIa[...,2] += R_0

        X_x_range = self.get_X_x_range(delta_x)
        X_phi_range = self.get_X_phi_range(delta_phi, R_0)
        n_idx_x = len(X_x_range)
        n_idx_phi = len(X_phi_range)
        idx_x = np.arange(n_idx_x)
        idx_phi = np.arange(n_idx_phi)

        idx_x_ic = idx_x[(n_idx_x) % 2::2]
        idx_x_id = idx_x[(n_idx_x + 1) % 2::2]
        idx_phi_ic = idx_phi[(n_idx_phi) % 2::2]
        idx_phi_id = idx_phi[(n_idx_phi + 1) % 2::2]

        X_E = X_x_range[idx_x_ic]
        X_F = X_x_range[idx_x_id]

        X_CIa = X_dIa[idx_phi_ic]
        X_DIa = X_dIa[idx_phi_id]

        expand = np.array([1,0,0])
        X_E_a = np.einsum('i,j->ij', X_E, expand)
        X_ECIa = X_CIa[np.newaxis,:,:,:] + X_E_a[:,np.newaxis,np.newaxis,:]
        X_F_a = np.einsum('i,j->ij', X_F, expand)
        X_FDIa = X_DIa[np.newaxis,:,:,:] + X_F_a[:,np.newaxis,np.newaxis,:]

        X_Ia = np.vstack([X_ECIa.flatten().reshape(-1,3), X_FDIa.flatten().reshape(-1,3)])
        return X_Ia

    I_cells_Fi = tr.Property(depends_on='+GEO')
    '''Array with nodal coordinates I - node, a - dimension
    '''
    @tr.cached_property
    def _get_I_cells_Fi(self):
        I_Fi_cell = self.wb_cell.I_Fi
        n_I_cell = self.wb_cell.n_I
        n_cells = self.n_cells
        i_range = np.arange(n_cells) * n_I_cell
        I_Fi = (I_Fi_cell[np.newaxis,:,:] + i_range[:, np.newaxis, np.newaxis]).reshape(-1,3)
        return I_Fi

    X_Ia = tr.Property(depends_on='+GEO')
    '''Array with nodal coordinates I - node, a - dimension
    '''
    @tr.cached_property
    def _get_X_Ia(self):
        idx_unique, idx_remap = self.unique_node_map
        return self.X_cells_Ia[idx_unique]

    I_Fi = tr.Property(depends_on='+GEO')
    '''Facet - node mapping
    '''
    @tr.cached_property
    def _get_I_Fi(self):
        _, idx_remap = self.unique_node_map
        return idx_remap[self.I_cells_Fi]

    node_match_threshold = tr.Property(depends_on='+GEO')

    def _get_node_match_threshold(self):
        min_length = np.min([self.a, self.b, self.c])
        return min_length * 1e-4

    unique_node_map = tr.Property(depends_on='+GEO')
    '''Property containing the mapping between the crease pattern nodes
    with duplicate nodes and pattern with compressed nodes array.
    The criterion for removing a node is geometric, the threshold
    is specified in node_match_threshold.
    '''
    def _get_unique_node_map(self):
        # reshape the coordinates in array of segments to the shape (n_N, n_D
        x_0 = self.X_cells_Ia
        # construct distance vectors between every pair of nodes
        x_x_0 = x_0[:, np.newaxis, :] - x_0[np.newaxis, :, :]
        # calculate the distance between every pair of nodes
        dist_0 = np.sqrt(np.einsum('...i,...i', x_x_0, x_x_0))
        # identify those at the same location
        zero_dist = dist_0 < self.node_match_threshold
        # get their indices
        i_idx, j_idx = np.where(zero_dist)
        # take only the upper triangle indices
        upper_triangle = i_idx < j_idx
        idx_multi, idx_delete = i_idx[upper_triangle], j_idx[upper_triangle]
        # construct a boolean array with True at valid and False at deleted
        # indices
        idx_unique = np.ones((len(x_0),), dtype='bool')
        idx_unique[idx_delete] = False
        # Boolean array of nodes to keep - includes both those that
        # are unique and redirection nodes to be substituted for duplicates
        idx_keep = np.ones((len(x_0),), dtype=np.bool_)
        idx_keep[idx_delete] = False
        # prepare the enumeration map map
        ij_map = np.ones_like(dist_0, dtype=np.int_) + len(x_0)
        i_ = np.arange(len(x_0))
        # indexes of nodes that are being kept
        idx_row = i_[idx_keep]
        # enumerate the kept nodes by putting their number onto the diagonal
        ij_map[idx_keep, idx_keep] = np.arange(len(idx_row))
        # broadcast the substitution nodes into the interaction positions
        ij_map[i_idx, j_idx] = ij_map[i_idx, i_idx]
        # get the substitution node by picking up the minimum index within ac column
        idx_remap = np.min(ij_map, axis=0)

        return idx_unique, idx_remap

    I_CDij = tr.Property(depends_on='+GEO')
    @tr.cached_property
    def _get_I_CDij(self):
        n_cells, n_ic, n_id, _, x_cell_idx, _, y_cell_idx = self.cell_map
        x_idx, y_idx = x_cell_idx / 2, y_cell_idx / 2
        n_x_, n_y_ = len(x_idx), len(y_idx)
        I_cell_offset = (n_ic + np.arange(n_x_ * n_y_).reshape(n_x_, n_y_)) * self.wb_cell.n_I
        I_CDij_map = (I_cell_offset.T[:, :, np.newaxis, np.newaxis] +
                      self.wb_cell.I_boundary[np.newaxis, np.newaxis, :, :])
        return I_CDij_map

    def setup_plot(self, pb):
        X_Ia = self.X_Ia.astype(np.float32)
        I_Fi = self.I_Fi.astype(np.uint32)

        I_M = self.I_CDij[(0, -1), :, (0, -1), :]
        _, idx_remap = self.unique_node_map
        J_M = idx_remap[I_M]
        X_Ma = X_Ia[J_M.flatten()]

        k3d_mesh = k3d.mesh(X_Ia,
                                 I_Fi,
                                 color=0x999999,
                                 side='double')

        pb.objects['k3d_mesh'] = k3d_mesh
        pb.plot_fig += k3d_mesh

        if self.show_nodes:
            self._add_nodes_to_fig(pb, X_Ma)

        if self.wb_cell.show_node_labels:
            self._add_nodes_labels_to_fig(pb, X_Ia)

        if self.show_wireframe:
            self._add_wireframe_to_fig(pb, X_Ia, I_Fi)

    def update_plot(self, pb):
        X_Ia = self.X_Ia.astype(np.float32)
        I_Fi = self.I_Fi.astype(np.uint32)

        I_M = self.I_CDij[(0, -1),:,(0, -1),:]
        _, idx_remap = self.unique_node_map
        J_M = idx_remap[I_M]
        X_Ma = X_Ia[J_M.flatten()]

        mesh = pb.objects['k3d_mesh']
        mesh.vertices = X_Ia
        mesh.indices = I_Fi

        if self.show_nodes:
            if self.NODES in pb.objects:
                pb.objects[self.NODES].positions = X_Ma
            else:
                self._add_nodes_to_fig(pb, X_Ma)
        else:
            if self.NODES in pb.objects:
                pb.clear_object(self.NODES)

        if self.show_wireframe:
            if self.WIREFRAME in pb.objects:
                wireframe = pb.objects[self.WIREFRAME]
                wireframe.vertices = X_Ia
                wireframe.indices = I_Fi
            else:
                self._add_wireframe_to_fig(pb, X_Ia, I_Fi)
        else:
            if self.WIREFRAME in pb.objects:
                pb.clear_object(self.WIREFRAME)

        if self.show_node_labels:
            if self.NODES_LABELS in pb.objects:
                pb.clear_object(self.NODES_LABELS)
            self._add_nodes_labels_to_fig(pb, X_Ia)
        else:
            if self.NODES_LABELS in pb.objects:
                pb.clear_object(self.NODES_LABELS)

    def _add_nodes_labels_to_fig(self, pb, X_Ia):
        text_list = []
        for I, X_a in enumerate(X_Ia):
            k3d_text = k3d.text('%g' % I, tuple(X_a), label_box=False, size=0.8, color=0x00FF00)
            pb.plot_fig += k3d_text
            text_list.append(k3d_text)
        pb.objects[self.NODES_LABELS] = text_list

    def _add_wireframe_to_fig(self, pb, X_Ia, I_Fi):
        k3d_mesh_wireframe = k3d.mesh(X_Ia,
                                      I_Fi,
                                      color=0x000000,
                                      wireframe=True)
        pb.plot_fig += k3d_mesh_wireframe
        pb.objects[self.WIREFRAME] = k3d_mesh_wireframe

    def _add_nodes_to_fig(self, pb, X_Ma):
        k3d_points = k3d.points(X_Ma, point_size=300)
        pb.objects[self.NODES] = k3d_points
        pb.plot_fig += k3d_points

    def _show_or_hide_fig_object(self, pb, show_obj, obj_name, obj_add_fun, obj_update_fun):
        if show_obj:
            if obj_name in pb.objects:
                obj_update_fun()
            else:
                obj_add_fun()
        else:
            if obj_name in pb.objects:
                pb.clear_object(obj_name)

    def export_fold_file(self, path=None):
        # See https://github.com/edemaine/fold/blob/master/doc/spec.md for fold file specification
        # Viewer: https://edemaine.github.io/fold/examples/foldviewer.html

        output_data = {
            "file_spec": 1,
            "file_creator": "BMCS software suite",
            "file_author": "RWTH Aachen - Institute of Structural Concrete",
            "file_title": "Preliminary Base",
            "file_classes": ["singleModel"],
            "frame_title": "Preliminary Base Crease Pattern",
            "frame_classes": ["creasePattern"],
            "vertices_coords": self.X_Ia.tolist(),
            "faces_vertices": self.I_Fi.tolist(),
            # To be completed
        }

        if path is None:
            path = time.strftime("%Y%m%d-%H%M%S") + '-shell.fold'

        with open(path, 'w') as outfile:
            json.dump(output_data, outfile, sort_keys=True, indent=4)
